""" Defines a set of features"""

from typing import Union

from numpy import nan
from pandas import DataFrame, Series

from ..utils.attributes import get_bool_attribute
from .qualitatives import (
    CategoricalFeature,
    OrdinalFeature,
    QualitativeFeature,
    get_categorical_features,
    get_ordinal_features,
)
from .quantitatives import QuantitativeFeature, get_quantitative_features
from .utils.base_feature import BaseFeature
from .utils.grouped_list import GroupedList

# class AutoFeatures(Features):
#     """TODO"""

#     __name__ = "AutoFeatures"

#     def __init__(self):
#         raise EnvironmentError(
#             f"[{self.__name__}] Should be instantiated with AutoFeatures.from_dataframe()"
#         )

#     def from_dataframe(self, X: DataFrame) -> None:
#         """Automatically generates Features from an input DataFrame based on there data types"""
#         # initiating features
#         categoricals, ordinals, quantitatives, datetimes = ([],) * 4

#         # getting data types
#         for feature, dtype in X.dtypes:
#             str_dtype = str(dtype).lower()
#             # categorical feature
#             if dtype == "object":
#                 categoricals += [feature]
#             # quantitative feature
#             elif str_dtype.startswith("int") or str_dtype.startswith("float"):
#                 quantitatives += [feature]
#             # datetime feature
#             elif "date" in str_dtype or "time" in str_dtype:
#                 datetimes += [feature]
#             # unknown data type
#             else:
#                 warn(
#                     f"[{self.__name__}] Ommited column {feature}, unknown data type {dtype}",
#                     UserWarning,
#                 )


def check_ordinal_features(
    ordinals: Union[list[OrdinalFeature], dict[str, list[str]]]
) -> list[OrdinalFeature]:
    """Checks that ordinals are correctly formatted"""

    # checking for ordinal types
    if isinstance(ordinals, (Features, list)):
        for feature in ordinals:
            if not isinstance(feature, OrdinalFeature):
                raise TypeError(
                    "Ordinals should be a list of OrdinalFeature or a dict of ordinal values."
                )
        ordinal_features = ordinals
    elif isinstance(ordinals, dict):
        ordinal_features = list(ordinals.keys())
    else:
        raise TypeError("Ordinals should be a list of OrdinalFeature or a dict of ordinal values.")

    return ordinal_features


class Features:
    """A set of typed features"""

    __name__ = "Features"

    def __init__(
        self,
        categoricals: list[Union[CategoricalFeature, str]] = None,
        quantitatives: list[Union[QuantitativeFeature, str]] = None,
        ordinals: Union[list[OrdinalFeature], dict[str, list[str]]] = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------

        categoricals : list[Union[CategoricalFeature, str]], optional
            List of categorical features or column names, by default ``None``

        quantitatives : list[Union[QuantitativeFeature, str]], optional
            List of quantitative features or column names, by default ``None``

        ordinals : Union[list[OrdinalFeature], dict[str, list[str]]], optional
            List of ordinal features or dict column names with associated value ordering,
            by default ``None``


        .. warning::
            At least one of categoricals, ordinals or quantitatives should be provided.


        Keyword Arguments
        -----------------

        ordinal_encoding : bool, optional
            Whether or not to ordinal encode labels, by default ``False``

        nan : str, optional
            Label for missing values, by default ``"__NAN__"``

        default : str, optional
            Label for default values, by default ``"__OTHER__"``
        """
        # initiating ordinal values if not provided
        if ordinals is None:
            ordinals = {}

        # getting list of ordinal features by name of BaseFeature
        ordinal_features = check_ordinal_features(ordinals)

        # casting features accordingly
        all_features = cast_features(categoricals, CategoricalFeature, **kwargs)
        all_features += cast_features(quantitatives, QuantitativeFeature, **kwargs)
        all_features += cast_features(
            ordinal_features, OrdinalFeature, ordinal_values=ordinals, **kwargs
        )

        # ensuring features are grouped accordingly (already initiated features)
        self._categoricals = get_categorical_features(all_features)
        self._ordinals = get_ordinal_features(all_features)
        self._quantitatives = get_quantitative_features(all_features)

        # checking that features were passed as input
        if len(self.categoricals) == 0 and len(self.quantitatives) == 0 and len(self.ordinals) == 0:
            raise ValueError(
                f"[{self}] No feature passed as input. Please provide column names"
                " by setting categoricals, quantitatives or ordinals."
            )

        # checking that qualitatitve and quantitative features are distinct
        check_duplicate_features(self.categoricals, self.quantitatives, self.ordinals)

        self._dropna = False
        self._ordinal_encoding = False
        self.is_fitted = get_bool_attribute(kwargs, "is_fitted", False)

    def __repr__(self) -> str:
        """Returns names of all features"""
        return f"{self.__name__}({str(self.versions)})"

    def __contains__(self, feature: Union[BaseFeature, str]) -> bool:
        """checks if a feature is in the features"""
        if isinstance(feature, BaseFeature):
            return feature.version in self.versions
        return feature in self.versions

    def __call__(self, feature_name: str) -> BaseFeature:
        """Returns specified feature by name"""

        # case for dataframes
        if isinstance(feature_name, DataFrame):
            return [feature.version for feature in self if feature.version in feature_name.columns]

        # looking for feature names
        self_dict = self.to_dict()
        if feature_name in self_dict:
            return self_dict.get(feature_name)

        # looking for version names
        if feature_name in self:
            return next(feature for feature in self if feature.version == feature_name)

        # not found feature
        raise ValueError(f"[{self.__name__}] '{feature_name}' not in features.")

    def __len__(self) -> int:
        """Returns number of features"""
        return len(self.to_list())

    def __iter__(self):
        """Returns an iterator of all features"""
        return iter(self.to_list())

    def __getitem__(
        self, index: Union[int, str, list[int], list[str], slice]
    ) -> Union[BaseFeature, list[BaseFeature]]:
        """Get item by index in list of features, by feature name or with a list of
        indices/feature names
        """
        # list index/slice request
        if isinstance(index, (int, slice)):
            return self.to_list()[index]

        # feature name request
        if isinstance(index, str):
            return self(index)

        # dataframe request
        if isinstance(index, DataFrame):
            index = list(index.columns)

        # list request and element to search for
        if isinstance(index, list) and len(index) > 0:
            # list index request
            if isinstance(index[0], int):
                self_list = self.to_list()
                return [self_list[idx] for idx in index]

            # feature name request
            if isinstance(index[0], str):
                return [self(name) for name in index]

        return None

    @property
    def names(self) -> list[str]:
        """Returns names of all features"""
        return get_names(self.to_list())

    @property
    def versions(self) -> list[str]:
        """Returns versions of all features"""
        return get_versions(self.to_list())

    @property
    def qualitatives(self) -> list[QualitativeFeature]:
        """Returns all qualitative features"""
        return self.categoricals + self.ordinals

    @property
    def categoricals(self) -> list[CategoricalFeature]:
        """Returns all categorical features"""
        return self._categoricals

    @categoricals.setter
    def categoricals(self, values: list[CategoricalFeature]) -> None:
        """sets ordinal features"""

        if not all(isinstance(feature, CategoricalFeature) for feature in values):
            raise AttributeError(
                f"[{self}] Trying to set categorical feature with wrongly typed feature"
            )
        self._categoricals = values

    @property
    def ordinals(self) -> list[OrdinalFeature]:
        """Returns all ordinal features"""
        return self._ordinals

    @ordinals.setter
    def ordinals(self, values: list[OrdinalFeature]) -> None:
        """sets ordinal features"""

        if not all(isinstance(feature, OrdinalFeature) for feature in values):
            raise AttributeError(
                f"[{self}] Trying to set ordinal feature with wrongly typed feature"
            )
        self._ordinals = values

    @property
    def quantitatives(self) -> list[QuantitativeFeature]:
        """Returns all quantitative features"""
        return self._quantitatives

    @quantitatives.setter
    def quantitatives(self, values: list[QuantitativeFeature]) -> None:
        """sets quantitative features"""

        if not all(isinstance(feature, QuantitativeFeature) for feature in values):
            raise AttributeError(
                f"[{self}] Trying to set quantitative feature with wrongly typed feature"
            )
        self._quantitatives = values

    @property
    def dropna(self) -> bool:
        """whether or not to drop missing values"""
        return self._dropna

    @dropna.setter
    def dropna(self, value: bool) -> None:
        """Sets features in dropna mode"""

        if not isinstance(value, bool):
            raise ValueError("Can only set dropna has a bool")

        for feature in self:  # iterating over each feature
            feature.dropna = value

        self._dropna = value

    @property
    def ordinal_encoding(self) -> bool:
        """whether or not to ordinal encode labels"""
        return self._ordinal_encoding

    @ordinal_encoding.setter
    def ordinal_encoding(self, value: bool) -> None:
        """Sets features in ordinal_encoding mode"""

        if not isinstance(value, bool):
            raise ValueError("Can only set ordinal_encoding has a bool")

        for feature in self:  # iterating over each feature
            feature.ordinal_encoding = value

        self._ordinal_encoding = value

    @property
    def content(self) -> dict:
        """Returns per feature content

        Returns
        -------
        dict
            per feature content
        """
        # returning all features' content
        return {feature.version: feature.content for feature in self}

    def remove(self, feature_version: str) -> None:
        """Removes a feature by version"""
        self.categoricals = remove_version(feature_version, self.categoricals)
        self.ordinals = remove_version(feature_version, self.ordinals)
        self.quantitatives = remove_version(feature_version, self.quantitatives)

    def keep(self, kept: list[str]) -> None:
        """list of features' versions to keep (removes the others)"""
        self.categoricals = keep_versions(kept, self.categoricals)
        self.ordinals = keep_versions(kept, self.ordinals)
        self.quantitatives = keep_versions(kept, self.quantitatives)

    def check_values(self, X: DataFrame) -> None:
        """Cheks for unexpected values for each feature in columns of DataFrame X"""
        # iterating over all features
        for feature in self:
            # checking for non-fitted features
            if not feature.is_fitted:
                raise RuntimeError(f"[{self.__name__}] '{feature}' not yet fitted!")

            # checking for unexpected values
            feature.check_values(X)

    def fit(self, X: DataFrame, y: Series = None) -> None:
        """fits all features to there respective column in DataFrame X"""
        # iterating over all features
        for feature in self:
            # checking for fitted features
            if feature.is_fitted:
                feature.check_values(X)

            # fitting feature
            else:
                feature.fit(X, y)

    def fillna(self, X: DataFrame, ignore_dropna: bool = False) -> DataFrame:
        """fills nans of a DataFrame"""

        # fills features with nans when dropna is True
        X.fillna(
            {
                feature.version: feature.nan
                for feature in self
                if feature.has_nan and (feature.dropna or ignore_dropna)
            },
            inplace=True,
        )

        return X

    def unfillna(self, X: DataFrame) -> DataFrame:
        """unfills nans when not supposed to have filled them"""

        # reinstating nans of features for which nans should not have been dropped
        X.replace(
            {
                feature.version: {feature.label_per_value.get(feature.nan, feature.nan): nan}
                for feature in self
                if feature.has_nan and not feature.dropna
            },
            inplace=True,
        )

        return X

    def update(
        self,
        feature_values: dict[str, GroupedList],
        convert_labels: bool = False,
        sorted_values: bool = False,
        replace: bool = False,
    ) -> None:
        """Updates all features using provided feature_values"""
        for feature, values in feature_values.items():  # updating each features
            self(feature).update(values, convert_labels, sorted_values, replace)

    def update_labels(self) -> None:
        """Updates all feature labels"""
        for feature in self:  # updating each features
            feature.update_labels()

    def add_feature_versions(self, y_classes: list[str]) -> None:
        """Builds versions of all features for each y_class"""
        self.categoricals = make_versions(self.categoricals, y_classes)
        self.ordinals = make_versions(self.ordinals, y_classes)
        self.quantitatives = make_versions(self.quantitatives, y_classes)

    def get_version_group(self, y_class: str) -> list[BaseFeature]:
        """Returns all features with specified version_tag"""

        return [feature for feature in self if feature.version_tag == y_class]

    @property
    def summary(self) -> DataFrame:
        """Summary of discretization process for all features"""
        # iterating over each feature
        summaries = []
        for feature in self:
            summaries += feature.summary

        # converting to DataFrame
        summaries = DataFrame(summaries)

        # defining indices to set
        indices = []
        for col in summaries.columns:
            if col not in ["feature", "label", "content", "target_mean", "frequency"]:
                indices += [col]
        indices = ["feature"] + indices + ["label"]

        return summaries.set_index(indices)

    def to_json(self, light_mode: bool = False) -> dict:
        """Serializes :class:`Features` for JSON saving

        Parameters
        ----------
        light_mode : bool, optional
            Whether or not to serialize in light mode (without statistics and history),
            by default ``False``

        """
        features_json = {feature.version: feature.to_json(light_mode) for feature in self}
        features_json.update({"is_fitted": self.is_fitted})
        return features_json

    def to_list(self) -> list[BaseFeature]:
        """Returns a list of all features"""
        return self.categoricals + self.ordinals + self.quantitatives

    def to_dict(self) -> dict[str, BaseFeature]:
        """Returns a dict of all versionned features"""
        return {feature.version: feature for feature in self.to_list()}

    @classmethod
    def load(cls, features_json: dict) -> "Features":
        """Allows one to load a set of :class:`Features`

        Parameters
        ----------
        features_json : dict
            Dictionary of serialized :class:`Features`

        Returns
        -------
        Features
            Loaded :class:`Features`.
        """

        # checking for fitted features
        is_fitted = features_json.pop("is_fitted", None)

        # casting each feature to there corresponding type
        unpacked_features: list[BaseFeature] = []
        for _, feature in features_json.items():
            # categorical feature
            if feature.get("is_categorical"):
                unpacked_features += [CategoricalFeature.load(feature)]

            # ordinal feature
            elif feature.get("is_ordinal"):
                unpacked_features += [OrdinalFeature.load(feature)]

            # ordinal feature
            elif feature.get("is_quantitative"):
                unpacked_features += [QuantitativeFeature.load(feature)]

        # initiating features
        return cls(unpacked_features, is_fitted=is_fitted)


def remove_version(removed_version: str, features: list[BaseFeature]) -> list[BaseFeature]:
    """removes a feature according its version"""
    return [feature for feature in features if feature.version != removed_version]


def keep_versions(kept_versions: list[str], features: list[BaseFeature]) -> list[BaseFeature]:
    """keeps requested feature versions according its version"""
    return [feature for feature in features if feature.version in kept_versions]


def make_versions(features: list[BaseFeature], y_classes: list[str]) -> BaseFeature:
    """Makes a copy of a list of features with specified version"""
    return [make_version(feature, y_class) for y_class in y_classes for feature in features]


def make_version(feature: BaseFeature, y_class: str) -> BaseFeature:
    """Makes a copy of a feature with specified version"""

    # converting feature to json
    feature_json = feature.to_json(light_mode=False)

    # categorical feature
    if feature_json.get("is_categorical"):
        new_feature = CategoricalFeature.load(feature_json)
    # ordinal feature
    elif feature_json.get("is_ordinal"):
        new_feature = OrdinalFeature.load(feature_json)
    # ordinal feature
    elif feature_json.get("is_quantitative"):
        new_feature = QuantitativeFeature.load(feature_json)
    # base feature
    else:
        new_feature = BaseFeature.load(feature_json)

    # modifying version and tag
    new_feature.version_tag = y_class
    new_feature.version = make_version_name(new_feature.name, y_class)

    return new_feature


def make_version_name(feature_name: str, y_class: str) -> str:
    """Builds a version name for a feature and target class"""

    return f"{feature_name}__y={y_class}"


def cast_features(
    features: list[str],
    target_class: type = BaseFeature,
    ordinal_values: dict[str, list[str]] = None,
    **kwargs,
) -> list[BaseFeature]:
    """converts a list of string feature names to there corresponding Feature class"""

    # inititating features if not provided
    if features is None:
        features: list[str] = []

    # inititating ordered values for ordinal features if not provided
    if ordinal_values is None:
        ordinal_values: dict[str, list[str]] = {}

    # initiating list of converted features
    converted_features: list[target_class] = []

    # iterating over each feature
    for feature in features:
        # string case, initiating feature
        if isinstance(feature, str):
            converted_features += [
                target_class(feature, values=ordinal_values.get(feature), **kwargs)
            ]
        # already a BaseFeature
        elif isinstance(feature, BaseFeature):
            converted_features += [feature]

        else:
            raise TypeError(
                f"[Features] feature {feature} is neither a str, nor a {target_class.__name__}."
            )

    # deduplicating features by version name
    return [
        feature
        for n, feature in enumerate(converted_features)
        if feature.version not in get_versions(converted_features[n + 1 :])
    ]


def get_names(features: list[BaseFeature]) -> list[str]:
    """Gives names from Features"""
    return [feature.name for feature in features]


def get_versions(features: list[BaseFeature]) -> list[str]:
    """Gives version names from Features"""
    return [feature.version for feature in features]


def check_duplicate_features(
    ordinals: list[OrdinalFeature],
    categoricals: list[CategoricalFeature],
    quantitatives: list[QuantitativeFeature],
) -> None:
    """Checks that features are distinct"""

    # getting feature names
    ordinal_names = get_versions(ordinals)
    categorcial_names = get_versions(categoricals)
    quantitative_names = get_versions(quantitatives)

    # checking for duplicates
    duplicate = [feature in ordinal_names + quantitative_names for feature in categorcial_names]
    if any(duplicate):
        raise ValueError(
            f"Provided categoricals found in ordinals/quantitatives: {duplicate}. "
            "Please, check inputs!"
        )
    duplicate = [feature in ordinal_names + categorcial_names for feature in quantitative_names]
    if any(duplicate):
        raise ValueError(
            f"Provided quantitatives found in ordinals/categoricals: {duplicate}. "
            "Please, check inputs!"
        )
    duplicate = [feature in quantitative_names + categorcial_names for feature in ordinal_names]
    if any(duplicate):
        raise ValueError(
            f"Provided ordinals found in categoricals/quantitatives: {duplicate}. "
            "Please, check inputs!"
        )



""" Measures of association between a Quantitative feature and binary target.
"""

from math import sqrt

from numpy import nan
from pandas import DataFrame, Series
from scipy.spatial.distance import correlation
from scipy.stats import kruskal, pearsonr, spearmanr
from statsmodels.formula.api import ols

from ...utils import extend_docstring
from .base_measures import AbsoluteMeasure, BaseMeasure, OutlierMeasure


class ReversibleMeasure(BaseMeasure):
    """A reversible measure"""

    is_reversible = True

    def __init__(self, threshold: float = 0.0) -> None:
        super().__init__(threshold)
        self.reversed = False

    def reverse_xy(self) -> bool:
        """reverses x and y within compute_association method"""
        self.reversed = True
        self.is_x_qualitative = not self.is_x_qualitative
        self.is_x_quantitative = not self.is_x_quantitative
        self.is_y_qualitative = not self.is_y_qualitative
        self.is_y_quantitative = not self.is_y_quantitative
        return True


class KruskalMeasure(ReversibleMeasure):
    """Kruskal-Wallis' test statistic between a Quantitative feature and a Qualitative target."""

    __name__ = "KruskalMeasure"
    is_x_quantitative = True
    is_y_qualitative = True

    @extend_docstring(BaseMeasure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # reversing if requested
        if self.reversed:
            x, y = y, x

        # ckecking for nans
        nans = x.isnull()

        # getting y values
        y_values = y.unique()

        # computing Kruskal-Wallis statistic
        kw = kruskal(*tuple(x[(~nans) & (y == y_value)] for y_value in y_values))
        self.value = kw[0] if kw else nan
        return self.value


class RMeasure(BaseMeasure):
    """Square root of the coefficient of determination of linear regression model of
    a Quantitative feature by a Binary target."""

    __name__ = "RMeasure"
    is_x_quantitative = True
    is_y_qualitative = True
    is_y_binary = True

    @extend_docstring(BaseMeasure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # # reversing if requested
        # if self.reversed:
        #     x, y = y, x

        # ckecking for nans
        nans = x.isnull() | x.isna()

        # checking values of y
        if len(y[~nans].unique()) != 2:
            raise ValueError(f"[{self}] Provided y is not binary")

        # grouping feature and target
        ols_df = DataFrame({"feature": x[~nans], "target": y[~nans]})

        # fitting regression of feature by target
        regression = ols("feature~C(target)", ols_df).fit()

        # computing R statistic
        self.value = (
            sqrt(regression.rsquared) if regression.rsquared and regression.rsquared >= 0 else nan
        )

        return self.value


class PearsonMeasure(AbsoluteMeasure):
    """Pearson's linear correlation coefficient between a Quantitative feature and target."""

    __name__ = "PearsonMeasure"

    is_x_quantitative = True
    is_y_quantitative = True

    @extend_docstring(BaseMeasure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # ckecking for nans
        nans = x.isnull() | x.isna()

        # computing spearman's r
        r = pearsonr(x[~nans], y[~nans])
        self.value = r[0] if r else nan
        return self.value


class SpearmanMeasure(AbsoluteMeasure):
    """Spearman's rank correlation coefficient between a Quantitative feature and target."""

    __name__ = "SpearmanMeasure"
    is_x_quantitative = True
    is_y_quantitative = True

    @extend_docstring(BaseMeasure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # ckecking for nans
        nans = x.isnull() | x.isna()
        # computing spearman's rho
        rho = spearmanr(x[~nans], y[~nans])
        self.value = rho[0] if rho else nan
        return self.value


class DistanceMeasure(AbsoluteMeasure):
    """Distance correlation between a Quantitative feature and target."""

    __name__ = "DistanceMeasure"
    is_x_quantitative = True
    is_y_quantitative = True

    @extend_docstring(BaseMeasure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # ckecking for nans
        nans = x.isnull()

        # computing distance correlation
        self.value = correlation(x[~nans], y[~nans])
        return self.value


class ZscoreOutlierMeasure(OutlierMeasure):
    """Z-Score based outlier measure"""

    __name__ = "ZScore"

    @extend_docstring(OutlierMeasure.compute_association)
    def compute_association(self, x: Series, y: Series = None) -> float:
        mean = x.mean()  # mean of the feature
        std = x.std()  # standard deviation of the feature
        zscore = (x - mean) / std  # zscore per observation

        # computing outlier rate
        outliers = abs(zscore) > 3
        self.value = outliers.mean()

        # keeping additional info
        self.info.update({"min": x.min(), "max": x.max(), "mean": mean, "std": std})

        return self.value


class IqrOutlierMeasure(OutlierMeasure):
    """Interquartile range based outlier measure"""

    __name__ = "IQR"

    @extend_docstring(OutlierMeasure.compute_association)
    def compute_association(self, x: Series, y: Series = None) -> float:
        q3 = x.quantile(0.75)  # 3rd quartile
        q1 = x.quantile(0.25)  # 1st quartile
        iqr = q3 - q1  # inter quartile range
        iqr_bounds = q1 - 1.5 * iqr, q3 + 1.5 * iqr  # bounds of the iqr range

        # computing outlier rate
        outliers = ~x.between(*iqr_bounds)
        self.value = outliers.mean()

        # keeping additional info
        self.info.update({"q1": q1, "median": x.median(), "q3": q3})

        return self.value



"""Sets of helper functions for BaseDiscretizer JSON serialization"""

from json import dumps, loads
from typing import Any, Union

from numpy import floating, inf, integer, isfinite

from ...config import Constants
from .grouped_list import GroupedList


def convert_value_to_base_type(value: Any) -> Union[str, float, int]:
    """Converts a value to python's base types (str, int, float) for JSON serialization.

    Parameters
    ----------
    value : Any
        A value to serialize

    Returns
    -------
    Union[str, float, int]
        Serialized value
    """

    output = value  # str/float/int value
    if not isinstance(value, str) and not isfinite(value):  # numpy.inf value
        output = Constants.INF
    elif isinstance(value, integer):  # np.int value
        output = int(value)
    elif isinstance(value, floating):  # np.float value
        output = float(value)

    return output


def convert_value_to_numpy_type(value: Union[str, float, int]) -> Any:
    """Converts a list or a dict of lists values to numpy types for JSON deserialization.

    Parameters
    ----------
    value : Union[str, float, int]
        A value to deserialize

    Returns
    -------
    Any
        Deserialized value
    """
    output = value  # str/float/int value
    if value == Constants.INF:  # numpy.inf value
        output = inf

    return output


def convert_values_to_base_types(
    iterable: Union[list[Any], dict[str, list[Any]]]
) -> Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]:
    """Converts a list or a dict of lists values to python's base types (str, int, float)
    for JSON serialization.

    Parameters
    ----------
    iterable : Union[list[Any], dict[str, list[Any]]]
        List or dict of lists of values to serialize

    Returns
    -------
    Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]
        List or dict of lists of values serialized
    """
    # list input
    output = None
    if isinstance(iterable, list):
        output = [convert_value_to_base_type(value) for value in iterable]
    # dict input
    elif isinstance(iterable, dict):
        output = {
            convert_value_to_base_type(key): convert_values_to_base_types(values)
            for key, values in iterable.items()
        }

    return output


def convert_values_to_numpy_types(
    iterable: Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]
) -> Union[list[Any], dict[str, list[Any]]]:
    """Converts a list or a dict of lists values to numpy types for JSON deserialization.

    Parameters
    ----------
    iterable : Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]
        List or dict of lists of values to deserialize

    Returns
    -------
    Union[list[Any], dict[str, list[Any]]]
        List or dict of lists of values deserialized
    """
    # list input
    output = None
    if isinstance(iterable, list):
        output = [convert_value_to_numpy_type(value) for value in iterable]
    # dict input
    elif isinstance(iterable, dict):
        output = {
            convert_value_to_numpy_type(key): convert_values_to_numpy_types(values)
            for key, values in iterable.items()
        }

    return output


def json_serialize_feature(feature: dict) -> str:
    """JSON serializes a feature's dict"""

    # serializing values
    values = feature.get("values")
    feature.update({"values": convert_values_to_base_types(values)})

    # serializing content
    content = feature.get("content")
    feature.update({"content": dumps(convert_values_to_base_types(content))})

    return feature


def json_serialize_combination(combination: dict) -> str:
    """JSON serializes a combination

    Parameters
    ----------
    combination : dict
        combination to serialize

    Returns
    -------
    dict
        JSON serialized combination
    """

    # converting combination to a json
    json_serialized_combination = {key: value for key, value in combination.items()}

    # converting combination values to a json
    json_serialized_combination.update(
        {
            "combination": [
                [convert_value_to_base_type(value) for value in modality]
                for modality in json_serialized_combination["combination"]
            ]
        }
    )

    return json_serialized_combination


def json_deserialize_content(json_serialized_feature: dict) -> dict:
    """JSON deserializes a content

    Parameters
    ----------
    json_serialized_content : str
        JSON serialized content

    Returns
    -------
    dict[str: GroupedList]
        content deserialized
    """

    # reading values and content from json
    values = json_serialized_feature.get("values")
    content = json_serialized_feature.get("content")

    # converting to numpy types
    values = convert_values_to_numpy_types(values)
    content = convert_values_to_numpy_types(loads(content))

    # cehcking for values
    if values is not None:
        # fixing json dumping with string keys
        for value in values:
            content_key = value

            # float and int dict keys (converted in string by json.dumps)
            if not isinstance(value, str) and isfinite(value):
                content_key = str(value)

            # updating
            content.update({value: content.pop(content_key)})

        # converting to grouped list
        return GroupedList(content)



""" Defines an ordinal feature"""

from ...utils import extend_docstring
from ..utils.base_feature import BaseFeature
from ..utils.grouped_list import GroupedList
from .qualitative_feature import QualitativeFeature


class OrdinalFeature(QualitativeFeature):
    """Defines an ordinal feature"""

    __name__ = "Ordinal"

    is_ordinal = True

    @extend_docstring(BaseFeature.__init__, append=False)
    def __init__(self, name: str, values: list[str], **kwargs) -> None:
        """
        Parameters
        ----------
        values : list[str]
            Ordered list of all unique values for the feature
        """
        super().__init__(name, **kwargs)

        # checking for values
        if values is None or len(values) == 0:
            raise ValueError(f"[{self}] Please make sure to provide values.")

        # checking for nan ordering
        if self.nan in values and not kwargs.get("load_mode", False):
            raise ValueError(
                f"[{self}] Ordering for '{self.nan}' can't be set by user, only fitted on data."
            )

        # checking for str values
        if not all(isinstance(value, str) for value in values):
            raise ValueError(f"[{self}] Please make sure to provide str values.")

        # saving up raw ordering for labeling
        self.raw_order = kwargs.get("raw_order", values[:])

        # setting values and labels
        super().update(GroupedList(values))

    def _specific_formatting(self, ordered_content: list[str]) -> str:
        """ordinal features' specific label formatting"""

        # ordered label for ordinal features
        return f"{ordered_content[0]} to {ordered_content[-1]}"


def get_ordinal_features(features: list[BaseFeature]) -> list[OrdinalFeature]:
    """returns ordinal features amongst provided features"""
    return [feature for feature in features if feature.is_ordinal]



"""Tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from pandas import DataFrame, Series, unique

from ...features import Features, QuantitativeFeature
from ...utils import extend_docstring
from ..qualitatives import OrdinalDiscretizer
from ..utils.base_discretizer import BaseDiscretizer, Sample
from .continuous_discretizer import ContinuousDiscretizer


class QuantitativeDiscretizer(BaseDiscretizer):
    """Automatic discretization pipeline of continuous and discrete features.

    Pipeline steps: :ref:`ContinuousDiscretizer`, :ref:`OrdinalDiscretizer`

    Modalities/values of features are grouped according to there respective orders:

     * [Continuous/Discrete features] real order of the values.
    """

    __name__ = "QuantitativeDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__, append=False, exclude=["features"])
    def __init__(
        self,
        quantitatives: list[QuantitativeFeature],
        min_freq: float,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------

        quantitatives : list[QuantitativeFeature]
            Quantitative features to process
        """

        # Initiating BaseDiscretizer
        super().__init__(quantitatives, **dict(kwargs, min_freq=min_freq))

    @property
    def half_min_freq(self) -> float:
        """Half of the minimal frequency of a quantile."""
        return self.min_freq / 2

    def _prepare_data(self, sample: Sample) -> Sample:  # pylint: disable=W0222
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``QuantitativeDiscretizer.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        # checking for binary target and copying X
        sample = super()._prepare_data(sample)

        # checking for quantitative columns
        check_quantitative_dtypes(sample.X, self.features.versions, self.__name__)

        return sample

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series) -> None:  # pylint: disable=W0222
        # verbose if requested
        self._log_if_verbose("------\n---")

        # checking data before bucketization
        sample = self._prepare_data(Sample(X, y))

        # fitting continuous features if any
        sample.X = self._fit_continuous(**sample)

        # fitting continuous features with rare modalities if any
        self._fit_continuous_with_rare_modalities(**sample)

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        if self.verbose:  # verbose if requested
            print("------\n")

        return self

    def _fit_continuous(self, X: DataFrame, y: Series) -> DataFrame:
        """Fit the ContinuousDiscretizer on the continuous features."""

        # [Quantitative features] Grouping values into quantiles
        continuous_discretizer = ContinuousDiscretizer(
            quantitatives=self.features.quantitatives,
            # copy needs to be True not to check for rare modalities
            **dict(self.kwargs, min_freq=self.min_freq, copy=True),
        )

        return continuous_discretizer.fit_transform(X, y)

    def _fit_continuous_with_rare_modalities(self, X: DataFrame, y: Series) -> None:
        """Fit the OrdinalDiscretizer on the continuous features with rare modalities."""

        # [Quantitative features] Grouping rare quantiles into closest common one
        #  -> can exist because of overrepresented values (values more frequent than min_freq)
        # identifying features that have rare modalities
        has_rare = check_frequencies(X, self.features, self.half_min_freq)

        # Grouping rare modalities
        if len(has_rare) > 0:
            ordinal_discretizer = OrdinalDiscretizer(
                ordinals=has_rare, **dict(self.kwargs, min_freq=self.half_min_freq, copy=False)
            )
            ordinal_discretizer.fit(X, y)


def check_frequencies(
    x: DataFrame, features: Features, half_min_freq: float
) -> list[QuantitativeFeature]:
    """Checks for rare modalities in the provided features."""

    # searching for features with rare quantiles: computing min frequency per feature
    frequencies = x[features.versions].apply(min_value_counts, features=features, axis=0)

    # identifying features that have rare modalities
    has_rare = list(frequencies[frequencies <= half_min_freq].index)

    # returning features with rare modalities
    return [feature for feature in features if feature.version in has_rare]


def min_value_counts(
    x: Series,
    features: Features,
    dropna: bool = False,
    normalize: bool = True,
) -> float:
    """Minimum of modalities' frequencies."""
    # getting corresponding feature
    feature = features(x.name)

    # modality frequency
    values = x.value_counts(dropna=dropna, normalize=normalize)

    # setting indices with known values
    if feature.values is not None:
        values = values.reindex(feature.labels).fillna(0)

    # minimal frequency
    return values.values.min()


def check_quantitative_dtypes(x: DataFrame, feature_versions: list[str], name: str) -> None:
    """Checks if the provided features are numeric."""

    # checking for numeric columns
    dtypes = x[feature_versions].map(type).apply(unique, result_type="reduce")

    # getting non-numeric columns
    not_numeric = dtypes.apply(lambda u: str in u)

    # raising error if non-numeric columns are found
    if any(not_numeric):
        raise ValueError(
            f"[{name}] Non-numeric features: "
            f"{str(list(not_numeric[not_numeric].index))} in provided quantitative_features. "
            "Please check your inputs."
        )



""" Defines a continuous/discrete feature"""

from .quantitative_feature import QuantitativeFeature


class NumericalFeature(QuantitativeFeature):
    """Defines a numerical feature"""

    __name__ = "Numerical"



""" Module for continuous combination evaluators. """

from abc import ABC

from pandas import Series
from scipy.stats import kruskal

from ..utils.combination_evaluator import AggregatedSample, CombinationEvaluator
from .continuous_target_rates import ContinuousTargetRate, TargetMean


class ContinuousCombinationEvaluator(CombinationEvaluator, ABC):
    """Continuous combination evaluator class."""

    is_y_continuous = True

    def _init_target_rate(self, target_rate: ContinuousTargetRate) -> None:
        """Initializes target rate."""
        if target_rate is None:
            self.target_rate = TargetMean()
        elif not isinstance(target_rate, ContinuousTargetRate):
            raise ValueError("target_rate must be a ContinuousTargetRate")
        else:
            self.target_rate = target_rate

    def _association_measure(
        self, xagg: AggregatedSample, n_obs: int = None, tol: float = 1e-10
    ) -> dict[str, float]:
        """Computes measures of association between feature and quantitative target.

        Parameters
        ----------
        xagg : DataFrame
            Values taken by y for each of x's modalities.

        Returns
        -------
        dict[str, float]
            Kruskal-Wallis' H as a dict.
        """
        _, _ = n_obs, tol  # unused attribute

        # Kruskal-Wallis' H
        return {"kruskal": kruskal(*tuple(xagg.values))[0]}

    def _grouper(self, xagg: AggregatedSample, groupby: dict[str:str]) -> Series:
        """Groups values of y

        Parameters
        ----------
        yval : Series
            _description_
        groupby : _type_
            _description_

        Returns
        -------
        Series
            _description_
        """
        # TODO: convert this to the vectorial version like BinaryCarver
        return xagg.groupby(groupby).sum()


class KruskalCombinations(ContinuousCombinationEvaluator):
    """Kruskal-Wallis' H based combination evaluation toolkit"""

    sort_by = "kruskal"



""" Defines a continuous feature"""

from numpy import diff, floor, inf, isfinite, log10, nan  # pylint: disable=E0611
from pandas import isna

from ..utils.base_feature import BaseFeature
from ..utils.grouped_list import GroupedList


class QuantitativeFeature(BaseFeature):
    """Defines a quantitative feature"""

    __name__ = "Quantitative"
    is_quantitative = True

    @BaseFeature.has_default.setter
    def has_default(self, value: bool) -> None:
        """does nothing for quantitative features: no default value possible"""
        _ = value

    def _specific_update(self, values: GroupedList, convert_labels: bool = False) -> None:
        """update content of values specifically per feature type"""

        # no values have been set
        if not convert_labels and self.values is None:
            # checking that inf is amongst values
            if values[-1] != inf:
                raise ValueError(f"[{self}] Must provide values with values[-1] == numpy.inf")
            self.values = values

        # values are not labels
        elif not convert_labels:
            # updating: iterating over each grouped values
            for kept_value, grouped_values in values.content.items():
                self.values.group(grouped_values, kept_value)

        # values are labels -> converting them back to values
        else:
            # iterating over each grouped values
            for kept_label, grouped_labels in values.content.items():
                # converting labels to values
                kept_value = self.value_per_label.get(kept_label)
                grouped_values = [self.value_per_label.get(label) for label in grouped_labels]

                # checking that kept values exists
                if kept_label not in self.value_per_label:
                    raise AttributeError(
                        f"{self} no {kept_label}, in value_per_label: {self.value_per_label}"
                    )

                # checking that grouped values exists
                for grouped_value, grouped_label in zip(grouped_values, grouped_labels):
                    if grouped_value is None:
                        print(
                            f"{self} no {grouped_label}, in value_per_label: {self.value_per_label}"
                        )

                # choosing which value to keep
                which_to_keep = [value for value in grouped_values if value != self.nan]

                # keeping the largest value amongst the discarded
                if len(which_to_keep) > 0:
                    kept_value = max(which_to_keep)

                # updating values if any to group
                if len(grouped_values) > 0:
                    # if ordinal_encoding, converting values to unique values
                    if self.ordinal_encoding:
                        r_value_per_label = {
                            v: self.values[k] for k, v in self.value_per_label.items()
                        }
                        grouped_values = [r_value_per_label[value] for value in grouped_values]
                        kept_value = r_value_per_label[kept_value]
                    self.values.group(grouped_values, kept_value)

                # updating statistics
                self._update_statistics_value(kept_label, kept_value)

    def make_labels(self) -> GroupedList:
        """gives labels per quantile (values for continuous features)

        Parameters
        ----------
        values : GroupedList
            feature's values (quantiles in the case of continuous ones)
        nan : str
            default string value for nan

        Returns
        -------
        list[str]
            list of labels per quantile
        """
        # filtering out nan and inf for formatting
        quantiles = [val for val in self.values if val != self.nan and isfinite(val)]

        # converting quantiles in string
        labels = format_quantiles(quantiles)

        # converting to grouped list
        labels = GroupedList(labels)

        # add NaNs if there are any (not grouped)
        if self.nan in self.values:
            labels.append(self.nan)

        # TODO add NaNs if there are any (grouped)
        # elif self.nan in self.values.values:

        return labels

    def _make_summary(self):
        """returns summary of feature's values' content"""
        # getting feature's labels
        labels = self.make_labels()

        # iterating over each value
        summary = []
        for num, (group, values) in enumerate(self.content.items()):
            # getting group label
            group_label = self.label_per_value.get(group)

            # Quantitative features: getting labels
            values = labels[num]

            # adding group summary
            summary += [{"feature": str(self), "label": group_label, "content": values}]

        # adding statistics and history
        return self._add_statistics_to_summary(summary)


def format_quantiles(a_list: list[float]) -> list[str]:
    """Formats a list of float quantiles into a list of boundaries.

    Rounds quantiles to the closest power of 1000.

    Parameters
    ----------
    a_list : list[float]
        Sorted list of quantiles to convert into string

    Returns
    -------
    list[str]
        List of boundaries per quantile
    """
    # only one non nan quantile
    if len(a_list) == 0:
        order = ["-inf < x < inf"]

    # several quantiles
    else:
        # getting minimal number of decimals to differentiate labels
        decimals_needed = min_decimals_to_differentiate(a_list, min_decimals=1)

        # scientific formatting
        formatted_list = [f"{number:.{decimals_needed}e}" for number in a_list]

        # stripping whitespaces
        formatted_list = [string.strip() for string in formatted_list]

        # low and high bounds per quantiles
        upper_bounds = formatted_list + [nan]
        lower_bounds = [nan] + formatted_list
        order: list[str] = []
        for lower, upper in zip(lower_bounds, upper_bounds):
            if isna(lower):
                order += [f"x <= {upper}"]
            elif isna(upper):
                order += [f"{lower} < x"]
            else:
                order += [f"{lower} < x <= {upper}"]

    return order


def min_decimals_to_differentiate(sorted_numbers: list[float], min_decimals: int = 0) -> int:
    """computes number of decimals needed for printing"""

    # checking for values
    if len(sorted_numbers) <= 1:
        return min_decimals

    # Find the smallest difference between consecutive numbers
    smallest_diff = min(diff(sorted_numbers))

    # All numbers are identical
    if smallest_diff == 0:
        return min_decimals

    # Number of decimal places needed
    decimal_places = -int(floor(log10(smallest_diff)))

    # minimum of 0
    return max(decimal_places, min_decimals) + 1


def get_quantitative_features(features: list[BaseFeature]) -> list[QuantitativeFeature]:
    """returns quantitative features amongst provided features"""
    return [feature for feature in features if feature.is_quantitative]



""" checks for installed extra dependencies"""


def has_idisplay() -> bool:
    """Returns True if the IPython.display module is available"""

    # trying to import extra dependencies
    try:
        from IPython.display import display_html

        _ = display_html
    except ImportError:
        _has_idisplay = False
    else:
        # trying to import other dependencies needed
        try:
            import ipykernel
            import jinja2
            import matplotlib

            _, _, _ = ipykernel, jinja2, matplotlib
        except ImportError:
            _has_idisplay = False
            print(
                "WARNING: IPython.display is available, but other dependencies are missing. "
                "Please install them using pip install AutoCarver[jupyter]."
            )
        else:
            _has_idisplay = True

    return _has_idisplay



""" Measures of association between a Qualitative feature and binary target.
"""

from math import sqrt

from pandas import Series, crosstab, notna
from scipy.stats import chi2_contingency

from ...utils import extend_docstring
from .base_measures import BaseMeasure

# X continue y continue distance correlation
# X binaire y continue kruskal y, x
# X multiclass y continue kruskal y, x

# X continue y binaire kruskal x, y
# X binaire y binaire cramerv/tschurpowt x, y
# X multiclass y binaire cramerv/tschurpowt x, y


# X continue y multiclass kruskal x, y
# X binaire y multiclass cramerv/tschurpowt x, y
# X multiclass y multiclass cramerv/tschurpowt x, y


class Chi2Measure(BaseMeasure):
    """Wrapper for `scipy.stats.chi2_contingency <https://docs.scipy.org/doc/scipy/reference/
    generated/scipy.stats.chi2_contingency.html>`_.
    Computes Chi2 statistic on the ``x`` by ``y`` `pandas.crosstab <https://pandas.pydata.org/docs/
    reference/api/pandas.crosstab.html>`_.
    """

    __name__ = "Chi2"
    is_x_qualitative = True
    is_y_qualitative = True

    @extend_docstring(BaseMeasure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # computing crosstab between x and y
        xtab = crosstab(x, y)

        # computing Chi2 statistic
        self.value = chi2_contingency(xtab)[0]
        return self.value


class CramervMeasure(Chi2Measure):
    """Computes Carmér's V between a Qualitative feature and a binary target."""

    __name__ = "CramervMeasure"

    @extend_docstring(Chi2Measure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # computing Chi2 if not provided
        chi2_value = super().compute_association(x, y)

        # number of non-missing observations
        n_obs = (notna(x) & notna(y)).sum()

        # number of values taken by the features
        n_mod_x, n_mod_y = x.nunique(), y.nunique()
        min_n_mod = min(n_mod_x, n_mod_y)

        # computing Cramér's V
        if min_n_mod > 1:
            self.value = sqrt(chi2_value / n_obs / (min_n_mod - 1))

        return self.value


class TschuprowtMeasure(Chi2Measure):
    """Computes Tschuprow's T between a Qualitative feature and a binary target."""

    __name__ = "TschuprowtMeasure"

    @extend_docstring(Chi2Measure.compute_association)
    def compute_association(self, x: Series, y: Series) -> float:
        # computing Chi2 if not provided
        chi2_value = super().compute_association(x, y)

        # number of non-missing observations
        n_obs = (notna(x) & notna(y)).sum()

        # number of values taken by the features
        n_mod_x, n_mod_y = x.nunique(), y.nunique()

        # computing Tschuprow's T
        dof_mods = sqrt((n_mod_x - 1) * (n_mod_y - 1))
        self.value = 0
        if dof_mods > 0:
            self.value = sqrt(chi2_value / n_obs / dof_mods)
        return self.value



"""Tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from pandas import DataFrame, Series

from ...features import QualitativeFeature
from ...utils import extend_docstring
from ..utils.base_discretizer import BaseDiscretizer, Sample
from .categorical_discretizer import CategoricalDiscretizer
from .chained_discretizer import check_frequencies, ensure_qualitative_dtypes
from .ordinal_discretizer import OrdinalDiscretizer


class QualitativeDiscretizer(BaseDiscretizer):
    """Automatic discretiziation pipeline of categorical and ordinal features.

    Pipeline steps: :ref:`CategoricalDiscretizer`, :ref:`StringDiscretizer`,
    :ref:`OrdinalDiscretizer`.

    Modalities/values of features are grouped according to there respective orders:

    * [Categorical features] order based on modality target rate.
    * [Ordinal features] user-specified order.
    """

    __name__ = "QualitativeDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__, append=False, exclude=["features"])
    def __init__(
        self,
        qualitatives: list[QualitativeFeature],
        min_freq: float,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------

        qualitatives : list[QualitativeFeature]
            Qualitative features to process

        """
        # Initiating BaseDiscretizer
        super().__init__(qualitatives, **dict(kwargs, min_freq=min_freq))

    def _prepare_data(self, sample: Sample) -> Sample:
        """Validates format and content of X and y. Converts non-string columns into strings."""
        sample.X = super()._prepare_X(sample.X)

        # checking feature values' frequencies
        check_frequencies(self.features, sample.X, self.min_freq, self.__name__)

        # converting non-str columns
        sample.X = ensure_qualitative_dtypes(self.features, sample.X, **self.kwargs)

        return sample

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series) -> None:  # pylint: disable=W0222
        # verbose if requested
        self._log_if_verbose("------\n---")

        # checking data before bucketization
        sample = self._prepare_data(Sample(X, y))

        # Base discretization (useful if already discretized)
        sample.X = self._base_transform(**sample)

        # fitting ordinal features if any
        self._fit_ordinals(**sample)

        # fitting categorical features if any
        self._fit_categoricals(**sample)

        # discretizing features based on each feature's values_order
        super().fit(**sample)

        if self.verbose:  # verbose if requested
            print("------\n")

        return self

    def _base_transform(self, X: DataFrame, y: Series) -> DataFrame:
        """Transform the data based on the previously fitted discretizers."""

        # looking for already discretized features
        discretized_features = [feature for feature in self.features if feature.is_fitted]

        # Base discretization (useful if already discretized)
        if len(discretized_features) > 0:
            base_discretizer = BaseDiscretizer(
                features=discretized_features, **dict(self.kwargs, copy=True, dropna=False)
            )
            X = base_discretizer.fit_transform(X, y)

        return X

    def _fit_ordinals(self, X: DataFrame, y: Series) -> None:
        """Fit the OrdinalDiscretizer on the ordinal features."""

        # [Qualitative ordinal features] Grouping rare values into closest common one
        if len(self.features.ordinals) > 0:
            ordinal_discretizer = OrdinalDiscretizer(
                ordinals=self.features.ordinals,
                **dict(self.kwargs, min_freq=self.min_freq, copy=False),
            )
            ordinal_discretizer.fit(X, y)

    def _fit_categoricals(self, X: DataFrame, y: Series) -> None:
        """Fit the CategoricalDiscretizer on the categorical features."""

        # [Qualitative non-ordinal features] Grouping rare values into default '__OTHER__'
        if len(self.features.categoricals) > 0:
            categorical_discretizer = CategoricalDiscretizer(
                categoricals=self.features.categoricals,
                **dict(self.kwargs, min_freq=self.min_freq, copy=False),
            )
            categorical_discretizer.fit(X, y)



""" Filters based on association measures between Qualitative features.
"""

from pandas import DataFrame

from ...features import BaseFeature, get_versions
from ...utils.extend_docstring import extend_docstring
from ..measures import CramervMeasure, TschuprowtMeasure
from .base_filters import BaseFilter


class QualitativeFilter(BaseFilter):
    """Computes max association between X and X (qualitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).
    """

    __name__ = "QualitativeFilter"

    is_x_qualitative = True

    @extend_docstring(BaseFilter.filter)
    def filter(self, X: DataFrame, ranks: list[BaseFeature]) -> list[BaseFeature]:
        # filtering ranks to avoid correlation with already removed features
        filtered_ranks = ranks[:]

        # iterating over each feature by target association order
        filtered: list[BaseFeature] = []
        for feature in ranks:
            # maximum correlation with a better feature
            correlation_with, worst_correlation = self._compute_worst_correlation(
                X, feature, filtered_ranks
            )

            # checking for too much correlation
            valid = self._validate(worst_correlation)

            # update feature accordingly (update stats)
            self._update_feature(
                feature, worst_correlation, valid, info={"correlation_with": correlation_with}
            )

            # keeping feature
            if valid:
                filtered += [feature]

            # removing feature from ranks
            else:
                filtered_ranks.remove(feature)

        return filtered

    def _compute_worst_correlation(
        self, X: DataFrame, feature: BaseFeature, rank: list[BaseFeature]
    ) -> tuple[str, float]:
        """Computes maximum association between a feature and features
        more associated to the target (according to ranks)
        """

        # initiating worst correlation
        correlation_with, worst_correlation = "itself", 0.0

        # features more associated with target
        current_feature_index = get_versions(rank).index(feature.version)
        better_features = rank[:current_feature_index]

        # iterating over each better feature
        for better_feature in better_features:
            # computing association with the better feature
            correlation = self.measure.compute_association(
                X[feature.version], X[better_feature.version]
            )

            # updating association if it's greater than previous better features
            if correlation > worst_correlation:
                worst_correlation, correlation_with = correlation, better_feature.version

            # breaking loop if too correlated
            if not self._validate(worst_correlation):
                break

        return correlation_with, worst_correlation

    def _validate(self, worst_correlation: float) -> bool:
        """Checks if the worst correlation of a feature is above specified threshold"""
        # dropping the feature if it was too correlated to a better feature
        valid = True
        if worst_correlation > self.threshold:
            valid = False

        return valid


class CramervFilter(QualitativeFilter):
    """Computes maximum Cramer's V between qualitative features of ``X``"""

    __name__ = "CramervFilter"

    @extend_docstring(QualitativeFilter.filter)
    def __init__(self, threshold: float = 1.0) -> None:
        super().__init__(threshold)
        self.measure = CramervMeasure(threshold)


class TschuprowtFilter(QualitativeFilter):
    """Computes maximum Tschuprow's T between qualitative features of ``X``"""

    __name__ = "TschuprowtFilter"

    @extend_docstring(QualitativeFilter.filter)
    def __init__(self, threshold: float = 1.0) -> None:
        super().__init__(threshold)
        self.measure = TschuprowtMeasure(threshold)



"""Binary combination evaluators package."""

from .binary_combination_evaluators import (
    BinaryCombinationEvaluator,
    CramervCombinations,
    TschuprowtCombinations,
)

__all__ = ["BinaryCombinationEvaluator", "TschuprowtCombinations", "CramervCombinations"]



""" Set of tools used for multiprocessing"""

from functools import partial
from multiprocessing import Pool
from typing import Callable

from pandas import DataFrame

from ...features import BaseFeature


def imap_unordered_function(fun: Callable, features: list[str], n_jobs: int, **kwargs):
    """converts a function to a multiprocessing imap_unordered format or list comprehension"""

    # initiating list of function results
    results = []

    # no multiprocessing: list comprehension
    if n_jobs <= 1:
        results = [fun(f, **kwargs) for f in features]
    # multiprocessing
    else:
        with Pool(processes=n_jobs) as pool:
            # feature processing
            results += pool.imap_unordered(partial(fun, **kwargs), features)

    return results


def apply_async_function(
    fun: Callable, features: list[BaseFeature], n_jobs: int, X: DataFrame, *args: list
):
    """converts a function to a multiprocessing apply_async format or list comprehension"""

    # no multiprocessing
    if n_jobs <= 1:
        results = [fun(f, X[f.version], *args) for f in features]

    # asynchronous transform of each feature
    else:
        with Pool(processes=n_jobs) as pool:
            results_async = [pool.apply_async(fun, (f, X[f.version], *args)) for f in features]

            #  waiting for the results
            results = [result.get() for result in results_async]

    return results



"""CombinationEvaluator class to evaluate
the best combination of modalities for a feature."""

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Union

from pandas import DataFrame, Series
from tqdm import tqdm

from ...features import BaseFeature, GroupedList
from ...utils import get_attribute, get_bool_attribute
from .combinations import (
    consecutive_combinations,
    format_combinations,
    nan_combinations,
    order_apply_combination,
    xagg_apply_combination,
)
from .target_rate import TargetRate
from .testing import TestKeys, is_viable, test_viability


@dataclass
class AggregatedSample:
    """Sample class to store aggregated samples

    Attributes
    ----------
    xagg : DataFrame
        Aggregated sample
    raw : DataFrame
        Raw aggregated sample
    """

    xagg: Union[DataFrame, Series]
    _raw: Union[DataFrame, Series] = None

    def __post_init__(self):
        """Post initialization"""
        # setting xtab_dev to xtab if not provided
        if self._raw is None and self.xagg is not None:
            self._raw = self.xagg.copy()

    @property
    def raw(self) -> DataFrame:
        """Returns the raw value of the xagg"""
        return self._raw

    @raw.setter
    def raw(self, value: Union[DataFrame, Series]) -> None:
        """Sets the raw value of the xagg"""

        # setting raw value
        self._raw = value

        # copying xagg
        if value is not None:
            self.xagg = value.copy()

    @property
    def shape(self) -> tuple[int, int]:
        """Returns the shape of the xagg"""
        return self.xagg.shape

    @property
    def index(self) -> list[str]:
        """Returns the index of the xagg"""
        return self.xagg.index

    @property
    def columns(self) -> list[str]:
        """Returns the columns of the xagg"""
        return self.xagg.columns

    @property
    def values(self) -> DataFrame:
        """Returns the values of the xagg"""
        return self.xagg.values

    def groupby(self, *args, **kwargs) -> DataFrame:
        """Groups the xagg by the specified indices"""
        return self.xagg.groupby(*args, **kwargs)


@dataclass
class AggregatedSamples:
    """stores train and dev samples"""

    train: AggregatedSample = field(default_factory=lambda: AggregatedSample(None))
    dev: AggregatedSample = field(default_factory=lambda: AggregatedSample(None))

    def set(self, train: DataFrame, dev: DataFrame = None) -> None:
        """Sets the train and dev samples"""

        # setting train and dev samples
        self.train = AggregatedSample(train)
        self.dev = AggregatedSample(dev)

    def dropna(self, feature_nan: str) -> None:
        """Removes nans from the samples"""

        # removing nans from crosstabs
        self.dev.xagg = filter_nan(self.dev.raw, feature_nan)
        self.train.xagg = filter_nan(self.train.raw, feature_nan)

    def set_indices_to_values(self, feature: BaseFeature) -> None:
        """Resets the index of the samples"""

        # renaming index to feature values
        if self.train.raw is not None:
            self.train.raw.rename(index=feature.value_per_label, inplace=True)
        if self.dev.raw is not None:
            self.dev.raw.rename(index=feature.value_per_label, inplace=True)

    def __repr__(self) -> str:
        return f"AggregatedSamples(train={self.train.shape}, dev={self.dev.shape})"

    def apply_combination(self, feature: BaseFeature) -> None:
        """Applies best combination to xagg"""

        # applying best_combination to xaggs
        self.train.raw = xagg_apply_combination(self.train.raw, feature)
        self.dev.raw = xagg_apply_combination(self.dev.raw, feature)


class CombinationEvaluator(ABC):
    """CombinationEvaluator class to evaluate
    the best combination of modalities for a feature."""

    __name__ = "CombinationEvaluator"

    is_y_binary = False
    is_y_continuous = False
    sort_by = None

    def __init__(
        self,
        max_n_mod: int = 5,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        max_n_mod : int, optional
            Maximum number of modalities per feature, by default ``5``

            * The combination with the best association will be selected.
            * All combinations of sizes from 1 to :attr:`max_n_mod` are tested out.

            .. tip::
                Set between ``3`` (faster, more robust) and ``7`` (slower, less robust)

        Keyword Arguments
        -----------------
        min_freq : float, optional
            Minimum frequency per modality per feature, by default ``None``

            * Features need at least one modality more frequent than :attr:`min_freq`
            * Defines number of quantiles of continuous features
            * Minimum frequency of modality of quantitative features

            .. tip::
                Set between ``0.01`` (slower, less robust) and ``0.2`` (faster, more robust)

        dropna : bool, optional
            * ``True``, try to group ``nan`` with other modalities.
            * ``False``, ``nan`` are ignored (not grouped), by default ``False``

        verbose : bool, optional
            * ``True``, without ``IPython``: prints raw statitics
            * ``True``, with ``IPython``: prints HTML statistics, by default ``False``
        """

        # setting attributes
        self.verbose = get_bool_attribute(kwargs, "verbose", False)
        self.dropna = get_bool_attribute(kwargs, "verbose", False)
        self.max_n_mod = max_n_mod
        self.min_freq = get_attribute(kwargs, "min_freq")

        # attributes to be set by get_best_combination
        self.feature: BaseFeature = None
        self.samples: AggregatedSamples = AggregatedSamples()
        self._statistics_cache = None
        self._init_target_rate(get_attribute(kwargs, "target_rate"))

    @abstractmethod
    def _init_target_rate(self, target_rate: TargetRate) -> None:
        """Initializes target rate."""
        self.target_rate = target_rate

    def __repr__(self) -> str:
        return f"{self.__name__}(max_n_mod={self.max_n_mod})"

    def _group_xagg_by_combinations(self, combinations: list[list]) -> list[dict]:
        """groups xagg by combinations of indices"""

        # values to groupby indices with
        indices_to_groupby = format_combinations(combinations)

        # grouping tab by its indices
        return [
            {
                "xagg": self._grouper(self.samples.train, index_to_groupby),
                "combination": combination,
                "index_to_groupby": index_to_groupby,
            }
            for combination, index_to_groupby in zip(
                combinations,
                tqdm(indices_to_groupby, desc="Grouping modalities   ", disable=not self.verbose),
            )
        ]

    def _compute_associations(self, grouped_xaggs: list[dict]) -> list[dict]:
        """computes associations for each grouped xagg"""

        # number of observations (only used for crosstabs)
        n_obs = self.samples.train.xagg.apply(sum).sum()

        # computing associations for each crosstab
        associations = [
            {**grouped_xagg, **self._association_measure(grouped_xagg["xagg"], n_obs=n_obs)}
            for grouped_xagg in tqdm(
                grouped_xaggs, desc="Computing associations", disable=not self.verbose
            )
        ]

        # sorting associations according to specified metric
        return (
            DataFrame(associations)
            .sort_values(self.sort_by, ascending=False)
            .to_dict(orient="records")
        )

    def _get_best_association(self, combinations: list[list[str]]) -> dict:
        """Computes associations of the tab for each combination

        Returns
        -------
        tuple[dict[str, Any], GroupedList]
            best viable association and associated modality order
        """
        # grouping tab by its indices
        grouped_xaggs = self._group_xagg_by_combinations(combinations)

        # computing associations for each tabs
        associations = self._compute_associations(grouped_xaggs)

        # testing viability of combination
        best_combination = self._get_viable_combination(associations)

        # applying best combination to feature labels and xtab
        self._apply_best_combination(best_combination)

        return best_combination

    def _apply_best_combination(self, best_association: dict) -> None:
        """Applies best combination to feature labels and xtab"""

        # checking that a combination was found
        if best_association is not None:
            # resetting samples' indices to values
            self.samples.set_indices_to_values(self.feature)

            # applying best_combination to feature labels
            labels = order_apply_combination(self.feature.labels, best_association["combination"])

            # updating feature's values according to combination
            self.feature.update(labels, convert_labels=True)

            # applying best_combination to raw xagg and xagg_dev
            self.samples.apply_combination(self.feature)

            # udpating statistics
            self.feature.statistics = self.target_rate.compute(self.samples.train.raw)

    def _get_best_combination_non_nan(self) -> dict:
        """Computes associations of the tab for each combination of non-nans

        - dropna has to be set to True
        """

        # raw ordering without nans
        raw_labels = GroupedList(self.feature.labels[:])

        # removing nans if any
        if self.feature.has_nan:
            # removing nans for combination of non-nans
            if self.feature.dropna:
                raw_labels.remove(self.feature.nan)

            # removing nans from crosstabs
            self.samples.dropna(self.feature.nan)

        # checking for non-nan values
        if self.samples.train.shape[0] > 1:
            # historizing raw combination
            self._historize_raw_combination()

            # all possible consecutive combinations
            combinations = consecutive_combinations(raw_labels, self.max_n_mod)

            # getting most associated combination
            return self._get_best_association(combinations)

        return None

    def _get_best_combination_with_nan(self, best_combination: dict) -> dict:
        """Computes associations of the tab for each combination with nans"""

        # grouping NaNs if requested to drop them (dropna=True)
        if self.dropna and self.feature.has_nan and best_combination is not None:
            # verbose if requested
            if self.verbose:
                print(f"[{self.__name__}] Grouping NaNs")

            # adding combinations with NaNs
            combinations = nan_combinations(self.feature, self.max_n_mod)

            # getting most associated combination
            best_combination = self._get_best_association(combinations)

        return best_combination

    def get_best_combination(
        self, feature: BaseFeature, xagg: DataFrame, xagg_dev: DataFrame = None
    ) -> dict:
        """Computes best combination of modalities for the feature"""

        # checking for min_freq
        if self.min_freq is None:
            raise ValueError("min_freq has to be set before calling get_best_combination")

        # setting dropna to user-requested value
        self.feature = feature
        self.feature.dropna = self.dropna

        # setting samples
        self.samples.set(train=xagg, dev=xagg_dev)

        # getting best combination without NaNs
        best_combination = self._get_best_combination_non_nan()

        # grouping NaNs if requested to drop them (dropna=True)
        return self._get_best_combination_with_nan(best_combination)

    def _test_viability_train(self, combination: dict) -> dict:
        """testing the viability of the combination on xagg_train"""

        # computing target rate and frequency per value
        train_rates = self.target_rate.compute(combination["xagg"])

        # viability on train sample:
        result = test_viability(train_rates, self.min_freq, self.target_rate.__name__)

        return result

    def _test_viability_dev(self, test_results: dict, combination: dict) -> dict:
        """testing the viability of the combination on xagg_dev"""

        # case 0: not viable on train or no test sample -> not testing for robustness
        if not test_results[TestKeys.VIABLE.value] or self.samples.dev.xagg is None:
            return {**test_results, "dev": {TestKeys.VIABLE.value: None}}

        # case 1: test sample provided and viable on train -> testing robustness
        # getting train target rates
        train_target_rate = test_results["train_rates"][self.target_rate.__name__]

        # grouping the dev sample per modality
        grouped_xagg_dev = self._grouper(self.samples.dev, combination["index_to_groupby"])

        # computing target rate and frequency per modality
        dev_rates = self.target_rate.compute(grouped_xagg_dev)

        # viability on dev sample:
        dev_results = test_viability(
            dev_rates, self.min_freq, self.target_rate.__name__, train_target_rate
        )
        test_results = {**test_results, **dev_results}

        # checking for viability on both samples
        test_results[TestKeys.VIABLE.value] = is_viable(test_results)

        return test_results

    def _get_viable_combination(self, associations: list[dict]) -> dict:
        """Tests the viability of all possible combinations onto xagg_dev"""

        # testing viability of all combinations
        viable_combination = None
        for n_combination, combination in tqdm(
            enumerate(associations),
            total=len(associations),
            desc="Testing robustness    ",
            disable=not self.verbose,
        ):
            # testing combination viability on train sample
            test_results = self._test_viability_train(combination)

            # testing combination viability on dev sample
            test_results = self._test_viability_dev(test_results, combination)

            # historizing combinations and tests
            self._historize_combination(combination, test_results)

            # best combination found: breaking the loop on combinations
            if test_results[TestKeys.VIABLE.value]:
                viable_combination = combination

                # historizing remaining combinations/not tested
                self._historize_remaining_combinations(associations, n_combination)
                break

        if self.verbose:  # verbose if requested
            print("\n")

        return viable_combination

    @abstractmethod
    def _grouper(self, xagg: AggregatedSample, groupby: dict[str, str]) -> Union[Series, DataFrame]:
        """Helper to group XAGG's values by groupby (carver specific)"""

    @abstractmethod
    def _association_measure(
        self, xagg: AggregatedSample, n_obs: int = None, tol: float = 1e-10
    ) -> dict[str, float]:
        """Helper to measure association between X and y (carver specific)"""

    def _historize_remaining_combinations(
        self, associations: list[dict], n_combination: int
    ) -> None:
        """historizes the remaining combinations that have not been tested"""

        # historizing all remaining combinations
        for combination in associations[n_combination + 1 :]:
            # historizing not tested combination
            self.feature.history = [
                {**clean_combination(combination, self.feature), "info": "Not checked"}
            ]

    def _historize_combination(self, combination: dict, test_results: dict) -> None:
        """historizes the test results of the combination"""

        # keeping only relevant information
        test_results.update(clean_combination(combination, self.feature))

        # checking for viability
        if test_results[TestKeys.VIABLE.value]:
            # setting feature's statistics (selected combination)
            # self.feature.statistics = test_results.pop("train_rates")
            self.feature.statistics = test_results.pop("train_rates")

            # adding info
            info = f"Best for {self.sort_by} and max_n_mod={self.max_n_mod}"

            # adding dropna info if its the case
            if test_results.get("dropna"):
                info += " (dropna=True)"

        # not viable on train or dev
        else:
            info = "Not viable"

            # removing train rates if still in there
            test_results.pop("train_rates", None)

        # saving info
        test_results.update({"info": info})

        # historizing test results
        self.feature.history = [test_results]

    def _historize_raw_combination(self):
        """historizes the raw combination"""

        # setting feature's statistics
        self.feature.statistics = self.target_rate.compute(self.samples.train.raw)

        # computing association of sample
        raw_association = self._association_measure(
            self.samples.train.raw, n_obs=sum(self.samples.train.raw.apply(sum))
        )

        # computing number of modalities
        n_mod = self.samples.train.raw.shape[0]

        # creating info message
        info = "Raw distribution"

        # adding info if n_mod > max_n_mod
        if n_mod > self.max_n_mod:
            info += f" (n_mod={n_mod}>max_n_mod={self.max_n_mod})"

        # historizing raw combination
        combination = {
            "info": info,
            **raw_association,
            "combination": {modality: modality for modality in self.samples.train.raw.index},
        }

        # historizing within feature
        self.feature.history = [{**combination, "n_mod": n_mod, "dropna": False}]

    def to_json(self) -> dict:
        """Converts to JSON format.

        To be used with ``json.dump``.

        Returns
        -------
        dict
            JSON serialized object
        """
        return {
            "sort_by": self.sort_by,
            "max_n_mod": self.max_n_mod,
            "dropna": self.dropna,
            "min_freq": self.min_freq,
            "verbose": self.verbose,
        }

    def save(self, file_name: str) -> None:
        """Saves :class:`CombinationEvaluator` to .json file.

        Parameters
        ----------
        file_name : str
            String of .json file name
        """
        # checking for input
        if file_name.endswith(".json"):
            with open(file_name, "w", encoding="utf-8") as json_file:
                json.dump(self.to_json(), json_file)
        # raising for non-json file name
        else:
            raise ValueError(f"[{self.__name__}] Provide a file_name that ends with .json.")

    @classmethod
    def load(cls, file: Union[str, dict]) -> "CombinationEvaluator":
        """Allows one to load a :class:`CombinationEvaluator` saved as a .json file.

        Parameters
        ----------
        file : str | dict
            String of .json file name or content of the file.

        Returns
        -------
        CombinationEvaluator
            A ready-to-use :class:`CombinationEvaluator`
        """
        # reading file
        if isinstance(file, str):
            with open(file, "r", encoding="utf-8") as json_file:
                combinations_json = json.load(json_file)
        elif isinstance(file, dict):
            combinations_json = file
        else:
            raise ValueError(f"[{cls.__name__}] Provide a file_name or a dict.")

        # checking for sort_by
        if combinations_json.get("sort_by") is None:
            if cls.sort_by is not None:
                raise ValueError(f"[{cls.__name__}] sort_by has to be {cls.sort_by}")
        elif combinations_json.get("sort_by") != cls.sort_by:
            raise ValueError(f"[{cls.__name__}] sort_by has to be {cls.sort_by}")

        # initiating BaseDiscretizer
        return cls(**combinations_json)


def filter_nan(xagg: Union[Series, DataFrame], str_nan: str) -> DataFrame:
    """Filters out nans from crosstab or y values"""

    # cehcking for values in crosstab
    filtered_xagg = None
    if xagg is not None:
        # filtering out nans if requested from train crosstab
        filtered_xagg = xagg.copy()
        if str_nan in xagg.index:
            filtered_xagg = xagg.drop(str_nan, axis=0)

    return filtered_xagg


def clean_combination(
    combination: dict, feature: BaseFeature, remove_train_rates: bool = False
) -> dict:
    """Cleans a combination to remove unwanted keys"""

    # removing unwanted keys
    index_to_groupby = combination.pop("index_to_groupby")

    # checking if dropna was used
    dropna = feature.nan in index_to_groupby

    # computing number of modalities
    n_mod = len(combination["combination"])

    # listing unwanted keys
    unwanted_keys = ["xagg", "combination"]
    if not remove_train_rates:
        unwanted_keys.append("train_rates")

    # filtering unwanted keys
    filtered = {k: v for k, v in combination.items() if k not in unwanted_keys}

    return {**filtered, "n_mod": n_mod, "dropna": dropna, "combination": index_to_groupby}



""" set of target rates for binary classification """

from abc import ABC

from numpy import log
from pandas import DataFrame, Series

from ..utils import TargetRate


class BinaryTargetRate(TargetRate, ABC):
    """Binary target rate class."""

    __name__ = "binary_target_rate"

    def compute(self, xagg: DataFrame) -> DataFrame:
        """Computes the target rate.

        Parameters
        ----------
        xagg : DataFrame
            A crosstab.

        Returns
        -------
        Series
            Target rate.
        """
        # checking for an xtab
        if xagg is not None:
            # frequency per modality
            frequency = xagg.sum(axis=1) / xagg.sum().sum()

            # computing target rate
            return DataFrame({self.__name__: self._compute(xagg), "frequency": frequency})
        return None


class TargetMean(BinaryTargetRate):
    """Mean target rate class."""

    __name__ = "target_mean"

    def _compute(self, xagg: DataFrame) -> Series:
        """Computes the mean target rate.

        Parameters
        ----------
        xagg : DataFrame
            A crosstab.

        Returns
        -------
        Series
            Mean target rate.
        """
        return xagg[1].divide(xagg.sum(axis=1))


class OddsRatio(TargetMean):
    """Odds ratio."""

    __name__ = "odds_ratio"

    def _compute(self, xagg: DataFrame) -> Series:
        """Computes the mean target rate.

        Parameters
        ----------
        xagg : DataFrame
            A crosstab.

        Returns
        -------
        Series
            Mean target rate.
        """
        target_rate = super()._compute(xagg)
        return target_rate / (1 - target_rate)


# class LogsOddsRatio(OddsRatio):
#     """Logs Odds ratio. same as WOE"""

#     __name__ = "log_odds_ratio"

#     def _compute(self, xagg: DataFrame) -> Series:
#         """Computes the mean target rate.

#         Parameters
#         ----------
#         xagg : DataFrame
#             A crosstab.

#         Returns
#         -------
#         Series
#             Mean target rate.
#         """
#         return log(super()._compute(xagg))


# class GiniCoefficient(BinaryTargetRate):
#     """Gini coefficient class."""

#     __name__ = "gini_coefficient"

#     def _compute(self, xagg: DataFrame) -> Series:
#         """Computes the Gini coefficient.

#         Parameters
#         ----------
#         xagg : DataFrame
#             A crosstab.

#         Returns
#         -------
#         Series
#             Gini coefficient.
#         """
#         sum_f = xagg.sum(axis=1)
#         squared = xagg.divide(sum_f, axis=0) ** 2
#         gini = 1 - squared.sum(axis=1)
#         return gini


class Woe(BinaryTargetRate):
    """Weight of evidence class."""

    __name__ = "woe"

    def _compute(self, xagg: DataFrame) -> Series:
        """Computes the Weight of evidence."""
        sum_f = xagg.sum(axis=1)
        means = xagg.divide(sum_f, axis=0)
        woe = log(means[1] / means[0])
        return woe


# class IV(Woe):
#     """Information Value coefficient class. TODO use for feature selection"""

#     __name__ = "iv"

#     def _compute(self, xagg: DataFrame) -> Series:
#         """Computes the Information Value ."""
#         sum_f = xagg.sum(axis=1)
#         means = xagg.divide(sum_f, axis=0)
#         woe = log(means[1] / means[0])
#         iv = (means[1] - means[0]) * woe
#         return iv



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for binary classification tasks.
"""

from numpy import unique
from pandas import DataFrame, Series, crosstab

from ..combinations import TschuprowtCombinations
from ..features import BaseFeature, Features
from ..utils import extend_docstring
from .utils.base_carver import BaseCarver, Samples


class BinaryCarver(BaseCarver):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a binary target.


    Examples
    --------
    `Binary Classification Example <https://autocarver.readthedocs.io/en/latest/examples/Carvers/
    BinaryClassification/binary_classification_example.html>`_
    """

    __name__ = "BinaryCarver"
    is_y_binary = True

    @extend_docstring(BaseCarver.__init__, exclude=["ordinal_encoding"])
    def __init__(
        self,
        features: Features,
        min_freq: float,
        dropna: bool = True,
        max_n_mod: int = 5,
        **kwargs,
    ) -> None:
        """
        Keyword Arguments
        -----------------

        ordinal_encoding : bool, optional
            Whether or not to ordinal encode :class:`Features`, by default ``True``

        combinations : BinaryCombinationEvaluator, optional
            Metric to perform association measure between :class:`Features` and target.

            .. tip::
                * Use :ref:`TschuprowtCombinations` for less, more robust, modalities
                * Use :ref:`CramervCombinations` for more, less robust, modalities
        """

        # default binary combinations
        combinations = kwargs.pop("combinations", None)
        if combinations is None:
            combinations = TschuprowtCombinations(max_n_mod=max_n_mod)

        # association measure used to find the best groups for binary targets
        if not combinations.is_y_binary:
            raise ValueError(
                f"[{self.__name__}] {combinations} is not suited for binary targets. "
                f"Choose from: TschuprowtCombinations, CramervCombinations."
            )

        # Initiating BaseCarver
        super().__init__(
            features=features,
            min_freq=min_freq,
            combinations=combinations,
            dropna=dropna,
            **kwargs,
        )

    def _prepare_data(self, samples: Samples) -> Samples:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``AutoCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated,
            by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame, dict[str, Callable]]
            Copies of (X, X_dev) and helpers to be used according to target type
        """
        # binary target, checking values
        y_values = unique(samples.train.y)
        if not ((0 in y_values) and (1 in y_values)) or len(y_values) != 2:
            raise ValueError(
                f"[{self.__name__}] y must be a binary Series of 0 and 1 (int or float, not object)"
            )

        # Checking for binary target and discretizing X
        return super()._prepare_data(samples)

    def _aggregator(self, X: DataFrame, y: Series) -> dict[str, DataFrame]:
        """Computes crosstabs for specified features and ensures that the crosstab is ordered
        according to the known labels

        Parameters
        ----------
        X : DataFrame
            _description_
        y : Series
            _description_

        Returns
        -------
        dict[str, DataFrame]
            dict of crosstab(X, y) by feature name
        """
        # checking for empty datasets (dev)
        xtabs = {feature.version: None for feature in self.features}
        if X is not None:
            # crosstab for each feature
            for feature in self.features:
                xtabs.update({feature.version: get_crosstab(X, y, feature)})

        return xtabs


def get_crosstab(X: DataFrame, y: Series, feature: BaseFeature) -> dict:
    """Computes crosstabs for specified features and ensures that the crosstab is ordered
    according to the known labels"""

    # computing crosstab between binary y and categorical X
    xtab = crosstab(X[feature.version], y)

    # reindexing to ensure the right order
    return xtab.reindex(feature.labels, fill_value=0)



""" Loads Quantitative Discretization tools."""

from .continuous_discretizer import ContinuousDiscretizer
from .quantitative_discretizer import QuantitativeDiscretizer

__all__ = [
    "QuantitativeDiscretizer",
    "ContinuousDiscretizer",
]



"""This module contains the continuous combinations module."""

from .continuous_combination_evaluators import ContinuousCombinationEvaluator, KruskalCombinations

__all__ = ["ContinuousCombinationEvaluator", "KruskalCombinations"]



"""Tools to build simple buckets out of Quantitative features
for a binary classification model.
"""

from numpy import array, digitize, in1d, inf, isnan, linspace, quantile, sort, unique
from pandas import DataFrame, Series

from ...features import GroupedList, QuantitativeFeature, get_versions
from ...utils import extend_docstring
from ..utils.base_discretizer import BaseDiscretizer
from ..utils.multiprocessing import imap_unordered_function


class ContinuousDiscretizer(BaseDiscretizer):
    """Automatic discretizing of continuous and discrete features, building simple groups of
    quantiles of values.

    Quantile discretization creates a lot of modalities (for example: up to 100 modalities for
    ``min_freq=0.01``).
    Set ``min_freq`` with caution.

    The number of quantiles depends on overrepresented modalities and nans:

    * Values more frequent than ``min_freq`` are set as there own modalities.
    * Other values are cut in quantiles using ``numpy.quantile``.
    * The number of quantiles is set as ``(1-freq_frequent_modals)/(min_freq)``.
    * Nans are considered as a modality (and are taken into account in ``freq_frequent_modals``).
    """

    __name__ = "ContinuousDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__, append=False, exclude=["features"])
    def __init__(
        self,
        quantitatives: list[QuantitativeFeature],
        min_freq: float,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------

        quantitatives : list[QuantitativeFeature]
            Quantitative features to process
        """

        # Initiating BaseDiscretizer
        super().__init__(features=quantitatives, **dict(kwargs, min_freq=min_freq))

    @property
    def q(self) -> int:
        """Number of quantiles to discretize the continuous features."""
        return round(1 / self.min_freq)

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series = None) -> None:  # pylint: disable=W0222
        self._log_if_verbose()  # verbose if requested

        # fitting each feature
        all_orders = imap_unordered_function(
            fit_feature,
            self.features.quantitatives,
            self.n_jobs,
            X=X[get_versions(self.features.quantitatives)],
            q=self.q,
        )

        # storing into the values_orders
        self.features.fit(X, y)
        self.features.update(dict(all_orders))

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self


def fit_feature(feature: QuantitativeFeature, X: DataFrame, q: float) -> tuple[str, GroupedList]:
    """Fits one feature"""

    # getting quantiles for specified feature
    quantiles = find_quantiles(X[feature.version].values, q=q)

    # Converting to a groupedlist
    order = GroupedList(quantiles + [inf])

    return feature.version, order


def find_quantiles(
    df_feature: array,
    q: int,
) -> list[float]:
    """Finds quantiles of a Series recursively.

    * Values more frequent than ``min_freq`` are set as there own modalities.
    * Other values are cut in quantiles using ``numpy.quantile``.
    * The number of quantiles is set as ``(1-freq_frequent_modals)/(min_freq)``.
    * Nans are considered as a modality (and are taken into account in ``freq_frequent_modals``).

    Parameters
    ----------
    df_feature : Series
        continuous feature
    q : int
        number of quantiles

    Returns
    -------
    list[float]
        list of quantiles for the feature
    """
    return list(
        sort(
            np_find_quantiles(
                df_feature[~isnan(df_feature)],  # getting rid of missing values
                q,
                initial_len_df=len(df_feature),  # getting raw dataset size
                quantiles=[],  # initiating list of quantiles
            )
        )
    )


def np_find_quantiles(
    df_feature: array,
    q: int,
    initial_len_df: int = None,
    quantiles: list[float] = None,
) -> list[float]:
    """Finds quantiles of a Series recursively.

    * Values more frequent than ``min_freq`` are set as there own modalities.
    * Other values are cut in quantiles using ``numpy.quantile``.
    * The number of quantiles is set as ``(1-freq_frequent_modals)/(min_freq)``.
    * Nans are considered as a modality (and are taken into account in ``freq_frequent_modals``).

    Parameters
    ----------
    df_feature : Series
        _description_
    q : int
        _description_
    initial_len_df : int, optional
        _description_, by default None
    quantiles : list[float], optional
        _description_, by default None

    Returns
    -------
    list[float]
        _description_
    """

    # case 1: no observation, all values have been attributed there corresponding modality
    if df_feature.shape[0] == 0:
        return quantiles

    # frequencies per known value
    values, frequencies = unique(df_feature, return_counts=True)

    # case 2 : there is an over-represented value
    if any(frequencies >= initial_len_df / q):
        # identifying over-represented modality
        frequent_values = values[frequencies >= initial_len_df / q]

        # computing quantiles on smaller and greater values
        sub_indices = digitize(df_feature, frequent_values, right=False)
        for i in range(0, len(frequent_values) + 1):
            quantiles += np_find_quantiles(
                df_feature[(sub_indices == i) & (~in1d(df_feature, frequent_values))],
                q,
                initial_len_df,
                [],
            )

        # adding over-represented modality to the list of quantiles
        return quantiles + list(frequent_values)

    # case 3 : there is no over-represented value -> computing quantiles
    quantiles += compute_quantiles(df_feature, q, initial_len_df)

    return quantiles


def compute_quantiles(df_feature: array, q: int, initial_len_df: int) -> list[float]:
    """Computes quantiles of a Series.

    - q should not be larger than len(df_feature)
    - df_feature should not contain any missing values
    - df_feature should not contain any over-represented modality
    - df_feature should not be empty
    """

    # getting quantiles needed
    quantiles_needed = get_remaining_quantiles(len(df_feature), initial_len_df, q)

    # cutting values into quantiles if there are enough values remaining
    if len(quantiles_needed) > 0:
        return list(quantile(df_feature, quantiles_needed, method="lower"))

    # -> not enough values remaining, grouping all remaining values into one quantile
    # returning the maximum value remaining, because it will be grouped within
    # QuantitativeDiscretizer when using OrdinalDiscretizer
    return [max(df_feature)]


def get_remaining_quantiles(remaining_len_df: array, initial_len_df: int, q: int) -> list[float]:
    """Computes list of indices of quantiles needed."""

    # updating number of quantiles taking into account identified over-represented modalities
    new_q = round(remaining_len_df / initial_len_df * q)

    # list of quantiles needed
    return linspace(0, 1, new_q + 1)[1:-1]



"""Tools to select the best Quantitative and Qualitative features for a Regression task."""

from .filters import (
    BaseFilter,
    NonDefaultValidFilter,
    SpearmanFilter,
    TschuprowtFilter,
    ValidFilter,
)
from .measures import BaseMeasure, KruskalMeasure, ModeMeasure, NanMeasure, SpearmanMeasure
from .utils.base_selector import BaseSelector


class RegressionSelector(BaseSelector):
    """A pipeline of measures to perform a feature pre-selection that maximizes association
    with a quantitative target.

    """

    __name__ = "RegressionSelector"

    def _initiate_measures(self, requested_measures: list[BaseMeasure] = None) -> list[BaseMeasure]:
        """initiates the list of measures with default values"""

        # initating to requested ones
        measures = requested_measures

        # if no measure were passed
        if measures is None:
            measures = [NanMeasure(), ModeMeasure(), SpearmanMeasure(), KruskalMeasure()]

        # adding default measure of mode
        mode_measure = ModeMeasure()
        if all(measure.__name__ != mode_measure.__name__ for measure in measures):
            measures = [mode_measure] + measures

        # adding default measure of nan
        nan_measure = NanMeasure()
        if all(measure.__name__ != nan_measure.__name__ for measure in measures):
            measures = [nan_measure] + measures

        # checking for measures for qualitative target
        if not all(
            measure.is_y_quantitative
            or (measure.reverse_xy() and measure.is_y_quantitative)
            or measure.is_default
            for measure in measures
        ):
            raise ValueError(f"[{self}] Provide measures for quantitative target!")
        return measures

    def _initiate_filters(self, requested_filters: list[BaseFilter] = None) -> list[BaseFilter]:
        """initiates the list of measures with default values"""

        # initating to requested ones
        filters = requested_filters

        # if no measure were passed
        if filters is None:
            filters = [ValidFilter(), TschuprowtFilter(), SpearmanFilter()]

        # adding default validity filter (based on measures)
        valid_filter = ValidFilter()
        if all(filter.__name__ != valid_filter.__name__ for filter in filters):
            filters = [valid_filter] + filters

        # adding default validity filter (based on measures)
        valid_filter = NonDefaultValidFilter()
        if all(filter.__name__ != valid_filter.__name__ for filter in filters):
            filters = [valid_filter] + filters

        return filters



""" Define the quantitative features. """

from .datetime_feature import DatetimeFeature
from .numerical_feature import NumericalFeature
from .quantitative_feature import QuantitativeFeature, get_quantitative_features

__all__ = [
    "NumericalFeature",
    "DatetimeFeature",
    "QuantitativeFeature",
    "get_quantitative_features",
]



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for any task.
"""

from typing import Any, Generator

from pandas import DataFrame

from ...features import BaseFeature, GroupedList


def combinations_at_index(
    start_index: int, elements: list[Any], remaining_groups: int
) -> Generator[list[Any], int, int]:
    """Gets all possible combinations of sizes up to the last element of a list"""

    # iterating over each possible length of groups
    for size in range(1, len(elements) + 1):
        next_index = start_index + size  # index from which to start the next group

        # checking that next index is not off the elements list
        if next_index < len(elements) + 1:
            # checking that there are remaining groups or that it is the last group
            if (remaining_groups > 1) | (next_index == len(elements)):
                combination = list(elements[start_index:next_index])
                yield (combination, next_index, remaining_groups - 1)


def consecutive_combinations(
    raw_order: list[Any],
    max_group_size: int,
    remaining_groups: int = None,
    current_combination: list[Any] = None,
    next_index: int = None,
    all_combinations: list[list[Any]] = None,
) -> list[list[Any]]:
    """Computes all possible combinations of values of order up to max_group_size."""

    # initiating recursive attributes
    if current_combination is None:
        current_combination = []
    if next_index is None:
        next_index = 0
    if remaining_groups is None:
        remaining_groups = max_group_size
    if all_combinations is None:
        all_combinations = []

    # getting combinations for next index
    next_combinations = list(combinations_at_index(next_index, raw_order, remaining_groups))

    # stop case: no next_combinations possible -> adding to all_combinations
    if len(next_combinations) == 0 and 1 < len(current_combination) <= max_group_size:
        # saving up combination
        all_combinations += [current_combination]

        # resetting remaining number of groups
        remaining_groups = max_group_size

    # otherwise: adding all next_combinations to the current_combination
    for combination, new_next_index, new_remaining_groups in next_combinations:
        # going a rank further in the raw_xtab
        consecutive_combinations(
            raw_order,
            max_group_size,
            remaining_groups=new_remaining_groups,
            current_combination=current_combination + [combination],
            next_index=new_next_index,
            all_combinations=all_combinations,
        )

    return all_combinations


def nan_combinations(
    feature: BaseFeature,
    max_n_mod: int,
) -> list[list[str]]:
    """All consecutive combinations of non-nans with added nan to each possible group and a last
    group only with nan if the max_n_mod is not reached by the combination

    - feature must have has_nan = True
    - feature must have dropna = True
    - len(feature.labels) <= max_n_mod
    """
    # raw ordering without nans
    raw_labels = GroupedList(feature.labels[:])
    raw_labels.remove(feature.nan)  # nans are added within nan_combinations

    # all possible consecutive combinations
    combinations = consecutive_combinations(raw_labels, max_n_mod)

    # iterating over each combination
    nan_combis = []
    for combination in combinations:
        # adding nan to each group of the combination
        nan_combination = []
        for n in range(len(combination)):
            # copying input combination
            new_combination = combination[:]
            # adding nan to the nth group
            new_combination[n] = new_combination[n] + [feature.nan]
            # storing updated combination with attributed group to nan
            nan_combination += [new_combination]

        # if max_n_mod is not reached adding a combination with nans alone
        if len(combination) < max_n_mod:
            # copying input combination
            new_combination = combination[:]
            # adding a group for nans only
            nan_combination += [new_combination + [[feature.nan]]]

        nan_combis += nan_combination

    # adding a combination for all modalities vs nans
    nan_combis += [[list(raw_labels), [feature.nan]]]

    return nan_combis


def order_apply_combination(order: GroupedList, combination: list[list[Any]]) -> GroupedList:
    """Converts a list of combination to a GroupedList"""
    order_copy = GroupedList(order)
    for combi in combination:
        order_copy.group(combi, combi[0])

    return order_copy


def xagg_apply_combination(xagg: DataFrame, feature: BaseFeature) -> DataFrame:
    """Applies an order (combination) to a crosstab

    Parameters
    ----------
    xagg : DataFrame
        Crosstab
    order : GroupedList
        Combination of index to apply to the crosstab

    Returns
    -------
    dict[str, Any]
        Ordered crosstab.
    """
    # checking for input values
    combi_xagg = None
    if xagg is not None:
        # grouping modalities in the crosstab
        groups = list(map(feature.label_per_value.get, xagg.index))
        combi_xagg = xagg.groupby(groups, dropna=False, sort=False).sum()

        # reindexing to ensure the right labels
        revised_index = feature.labels
        if feature.has_nan and feature.nan not in revised_index and not feature.dropna:
            revised_index += [feature.nan]

        combi_xagg.index = revised_index

    return combi_xagg


def combination_formatter(combination: list[list[str]]) -> dict[str, str]:
    """Attributes the first element of a group to all elements of a group"""
    return {modal: group[0] for group in combination for modal in group}


def format_combinations(combinations: list[list[str]]) -> list[dict[str, str]]:
    """Formats a list of combinations"""
    return [combination_formatter(combination) for combination in combinations]



""" Loads Discretization tools."""

from .discretizer import Discretizer
from .qualitatives import (
    CategoricalDiscretizer,
    ChainedDiscretizer,
    OrdinalDiscretizer,
    QualitativeDiscretizer,
)
from .quantitatives import ContinuousDiscretizer, QuantitativeDiscretizer
from .utils import BaseDiscretizer, Sample, StringDiscretizer

__all__ = [
    "BaseDiscretizer",
    "Sample",
    "Discretizer",
    "QualitativeDiscretizer",
    "QuantitativeDiscretizer",
    "CategoricalDiscretizer",
    "ChainedDiscretizer",
    "OrdinalDiscretizer",
    "ContinuousDiscretizer",
    "StringDiscretizer",
]



"""Tool to build optimized buckets out of Quantitative and Qualitative features
for any task.
"""

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Union

from pandas import DataFrame, Series

from ...combinations import (
    CombinationEvaluator,
    CramervCombinations,
    KruskalCombinations,
    TschuprowtCombinations,
)
from ...discretizers import BaseDiscretizer, Discretizer, Sample
from ...features import BaseFeature, Features, GroupedList
from ...utils import extend_docstring, get_attribute, get_bool_attribute, has_idisplay
from .pretty_print import index_mapper, prettier_xagg

# trying to import extra dependencies
_has_idisplay = has_idisplay()
if _has_idisplay:
    from IPython.display import display_html


@dataclass
class Samples:
    """
    A container for storing training and development samples.

    Attributes:
        train (Sample): The training sample, containing features (X) and target (y).
        dev (Sample): The development sample, containing features (X) and target (y).

    Example:
        >>> import pandas as pd
        >>> from base_carver import Sample, Samples
        >>> X_train = pd.DataFrame({"feature1": [1, 2, 3], "feature2": [4, 5, 6]})
        >>> y_train = pd.Series([0, 1, 0])
        >>> X_dev = pd.DataFrame({"feature1": [7, 8, 9], "feature2": [10, 11, 12]})
        >>> y_dev = pd.Series([1, 0, 1])
        >>> train_sample = Sample(X=X_train, y=y_train)
        >>> dev_sample = Sample(X=X_dev, y=y_dev)
        >>> samples = Samples(train=train_sample, dev=dev_sample)
        >>> print(samples.train.X)
           feature1  feature2
        0         1         4
        1         2         5
        2         3         6
        >>> print(samples.dev.y)
        0    1
        1    0
        2    1
        dtype: int64
    """

    train: Sample = field(default_factory=lambda: Sample(X=None))
    dev: Sample = field(default_factory=lambda: Sample(X=None))

    def fillna(self, features: Features) -> None:
        """fills up nans in X and X_dev"""
        self.train.X = features.fillna(self.train.X)
        if self.dev.X is not None:
            self.dev.X = features.fillna(self.dev.X)


class BaseCarver(BaseDiscretizer, ABC):
    """Automatic carving of continuous, discrete, categorical and ordinal
    features that maximizes association with a binary or continuous target.

    First fits a :class:`Discretizer`. Raw data should be provided as input (not a result of
    ``Discretizer.transform()``).
    """

    __name__ = "AutoCarver"
    is_y_binary = False
    is_y_continuous = False
    is_y_multiclass = False

    @extend_docstring(BaseDiscretizer.__init__)
    def __init__(
        self,
        features: Features,
        min_freq: float,
        combinations: CombinationEvaluator,
        dropna: bool = True,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------

        dropna : bool, optional
            * ``True``, try to group ``nan`` with other modalities.
            * ``False``, ``nan`` are ignored (not grouped), by default ``True``

        max_n_mod : int, optional
            Maximum number of modalities per feature, by default ``5``

            * The combination with the best association will be selected.
            * All combinations of sizes from 1 to :attr:`max_n_mod` are tested out.

            .. tip::
                Set between ``3`` (faster, more robust) and ``7`` (slower, less robust)

        Keyword Arguments
        -----------------

        discretizer_min_freq : float, optional
            Specific :attr:`min_freq` used by discretizers, by default ``None`` for ``min_freq/2``
        """

        # minimum frequency for discretizer
        self.discretizer_min_freq = get_attribute(kwargs, "discretizer_min_freq", min_freq / 2)

        # Initiating BaseDiscretizer
        super().__init__(
            features,
            **dict(
                kwargs,
                verbose=get_bool_attribute(kwargs, "verbose", False),
                ordinal_encoding=get_bool_attribute(kwargs, "ordinal_encoding", True),
                min_freq=min_freq,
                dropna=dropna,
                discretizer_min_freq=self.discretizer_min_freq,
            ),
        )

        # setting combinations evaluator
        self.combinations = combinations
        self.combinations.min_freq = self.min_freq
        self.combinations.verbose = self.verbose
        self.combinations.dropna = self.dropna

    @property
    def pretty_print(self) -> bool:
        """Returns the pretty_print attribute"""
        return self.verbose and _has_idisplay

    @property
    def max_n_mod(self) -> int:
        """Returns the max_n_mod attribute"""
        return self.combinations.max_n_mod

    @max_n_mod.setter
    def max_n_mod(self, value: int) -> None:
        """Sets the max_n_mod attribute"""
        self.combinations.max_n_mod = value

    def _prepare_data(self, samples: Samples) -> Samples:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``BaseCarver.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evalute the robustness of discretization, by default ``None``
            It should have the same distribution as X.

        y_dev : Series, optional
            Binary target feature with wich the robustness of discretization is evaluated,
            by default ``None``

        Returns
        -------
        tuple[DataFrame, DataFrame]
            Copies of (X, X_dev)
        """
        # checking for not provided y
        if samples.train.y is None:
            raise ValueError(f"[{self.__name__}] y must be provided, got {samples.train.y}")

        # Checking for binary target and copying X
        samples.train = super()._prepare_data(samples.train)
        samples.dev = super()._prepare_data(samples.dev)

        # discretizing features according to min_freq
        samples = discretize(self.features, samples, **self.kwargs)

        # setting dropna to True for filling up nans
        self.features.dropna = True

        # filling up nans
        samples.fillna(self.features)

        return samples

    def fit(  # pylint: disable=W0222
        self,
        X: DataFrame,
        y: Series,
        *,
        X_dev: DataFrame = None,
        y_dev: Series = None,
    ) -> None:
        """Finds the combination of modalities of X that provides the best association with y.
        If provided, X_dev set should be large enough to have the same distribution as X.

        Parameters
        ----------
        X : DataFrame
            Dataset to determine :class:`Features`' optimal carving.

        y : Series
            Target with wich the association is maximized.

        X_dev : DataFrame, optional
            Dataset to evaluate robustness of :class:`Features`, by default ``None``

        y_dev : Series, optional
            Target associated to ``X_dev``, by default ``None``
        """

        # checking for fitted features
        if self.features.is_fitted:
            raise ValueError(
                f"[{self.__name__}] features are already fitted or previous fit failed. "
                "Please reset your features."
            )

        # setting is_fitted
        self.features.is_fitted = True

        # initiating samples
        samples = Samples(Sample(X, y), Sample(X_dev, y_dev))

        # preparing datasets and checking for wrong values
        samples = self._prepare_data(samples)

        # logging if requested
        super()._log_if_verbose("---------\n------")

        # computing crosstabs for each feature on train/test
        xaggs = self._aggregator(**samples.train)
        xaggs_dev = self._aggregator(**samples.dev)

        # getting all features to carve (features are removed from self.features)
        all_features = self.features.versions

        # carving each feature
        for n, feature in enumerate(all_features):
            num_iter = f"{n+1}/{len(all_features)}"  # logging iteration number
            self._carve_feature(self.features(feature), xaggs, xaggs_dev, num_iter)

        # discretizing features based on each feature's values_order
        super().fit(X, y)

        return self

    @abstractmethod
    def _aggregator(self, X: DataFrame, y: Series) -> Union[Series, DataFrame]:
        """Helper that aggregates X by y into crosstab or means (carver specific)"""

    def _carve_feature(
        self,
        feature: BaseFeature,
        xaggs: dict[str, Union[Series, DataFrame]],
        xaggs_dev: dict[str, Union[Series, DataFrame]],
        num_iter: str,
    ) -> dict[str, GroupedList]:
        """Carves a feature into buckets that maximize association with the target"""

        # verbose if requested
        if self.verbose:
            print(f"--- [{self.__name__}] Fit {feature} ({num_iter})")

        # getting xtabs on train/test
        xagg = xaggs[feature.version]
        xagg_dev = xaggs_dev[feature.version]

        # printing raw distribution
        self._print_xagg(feature, xagg=xagg, xagg_dev=xagg_dev, message="Raw distribution")

        # getting best combination
        best_combination = self.combinations.get_best_combination(feature, xagg, xagg_dev=xagg_dev)

        # printing carved distribution, for found, suitable combination
        if best_combination is not None:
            self._print_xagg(
                feature,
                xagg=self.combinations.samples.train.xagg,
                xagg_dev=self.combinations.samples.dev.xagg,
                message="Carved distribution",
            )

        # no suitable combination has been found -> removing feature
        else:
            print(
                f"WARNING: No robust combination for {feature}. Consider increasing the size of "
                "X_dev or dropping the feature (X not representative of X_dev for this feature)."
            )
            self.features.remove(feature.version)

    def _print_xagg(
        self,
        feature: BaseFeature,
        xagg: Union[DataFrame, Series],
        message: str,
        *,
        xagg_dev: Union[DataFrame, Series] = None,
    ) -> None:
        """Prints crosstabs' target rates and frequencies per modality, in raw or html format

        Parameters
        ----------
        xagg : Union[DataFrame, Series]
            Train crosstab
        xagg_dev : Union[DataFrame, Series]
            Dev crosstab, by default None
        pretty_print : bool, optional
            Whether to output html or not, by default False
        """
        if self.verbose:
            print(f" [{self.__name__}] {message}")

            formatted_xagg, formatted_xagg_dev = self._format_xagg(feature, xagg, xagg_dev)

            nice_xagg, nice_xagg_dev = self._pretty_print(formatted_xagg, formatted_xagg_dev)

            if not self.pretty_print:  # no pretty hmtl printing
                self._print_raw(nice_xagg, nice_xagg_dev, xagg_dev)
            else:  # pretty html printing
                self._print_html(nice_xagg, nice_xagg_dev)

    def _format_xagg(
        self, feature: BaseFeature, xagg: DataFrame, xagg_dev: DataFrame = None
    ) -> tuple[DataFrame, DataFrame]:
        """Formats the XAGG DataFrame."""
        formatted_xagg = index_mapper(feature, xagg)
        formatted_xagg_dev = index_mapper(feature, xagg_dev)
        return formatted_xagg, formatted_xagg_dev

    def _pretty_print(
        self, formatted_xagg: DataFrame, formatted_xagg_dev: DataFrame
    ) -> tuple[str, str]:
        """Returns pretty-printed XAGG DataFrames."""
        nice_xagg = self.combinations.target_rate.compute(formatted_xagg)
        nice_xagg_dev = self.combinations.target_rate.compute(formatted_xagg_dev)
        return nice_xagg, nice_xagg_dev

    def _print_raw(self, nice_xagg: str, nice_xagg_dev: str, xagg_dev: DataFrame = None) -> None:
        """Prints raw XAGG DataFrames."""
        print(nice_xagg, "\n")
        if xagg_dev is not None:
            print("X_dev distribution\n", nice_xagg_dev, "\n")

    def _print_html(self, nice_xagg: str, nice_xagg_dev: str) -> None:
        """Prints XAGG DataFrames in HTML format."""
        # getting prettier xtabs
        nicer_xagg = prettier_xagg(nice_xagg, caption="X distribution")
        nicer_xagg_dev = prettier_xagg(nice_xagg_dev, caption="X_dev distribution", hide_index=True)

        # merging outputs
        nicer_xaggs = nicer_xagg + "          " + nicer_xagg_dev

        # displaying html of colored DataFrame
        display_html(nicer_xaggs, raw=True)

    @classmethod
    def load(cls, file_name: str) -> "BaseCarver":
        """Allows one to load a Carver saved as a .json file.

        The Carver has to be saved with ``Carver.save()``, otherwise there
        can be no guarantee for it to be restored.

        Parameters
        ----------
        file_name : str
            String of saved Carver's .json file name.

        Returns
        -------
        BaseDiscretizer
            A fitted Carver.
        """
        # reading file
        with open(file_name, "r", encoding="utf-8") as json_file:
            carver_json = json.load(json_file)

        # deserializing features
        features = Features.load(carver_json.pop("features"))

        # deserializing Combinations
        combinations = carver_json.pop("combinations")
        if combinations["sort_by"] == "tschuprowt":
            combinations = TschuprowtCombinations.load(combinations)
        elif combinations["sort_by"] == "cramerv":
            combinations = CramervCombinations.load(combinations)
        elif combinations["sort_by"] == "kruskal":
            combinations = KruskalCombinations.load(combinations)
        else:
            combinations = CombinationEvaluator.load(combinations)

        # initiating BaseDiscretizer
        return cls(features=features, combinations=combinations, **carver_json)


def discretize(
    features: Features,
    samples: Samples,
    discretizer_min_freq: float,
    **kwargs,
) -> Samples:
    """Discretizes X and X_dev according to the frequency of each feature's modalities."""

    # discretizing all features, always copying, to keep discretization from start to finish
    discretizer = Discretizer(
        features=features,
        **dict(
            kwargs,
            dropna=False,
            copy=True,
            ordinal_encoding=False,
            min_freq=discretizer_min_freq,
        ),
    )

    # fitting discretizer on X
    samples.train.X = discretizer.fit_transform(**samples.train)

    # applying discretizer on X_dev if provided
    if samples.dev.X is not None:
        samples.dev.X = discretizer.transform(**samples.dev)

    return samples



""" Loads Carving base tools."""

from .binary_carver import BinaryCarver
from .continuous_carver import ContinuousCarver
from .multiclass_carver import MulticlassCarver

__all__ = ["BinaryCarver", "ContinuousCarver", "MulticlassCarver"]



""" function to get boolean attributes from kwargs """


def get_attribute(kwargs: dict, attribute: str, default_value: any = None) -> any:
    """returns value from kwargs"""

    # getting attribute value
    value = kwargs.get(attribute)

    # attributing default value if not found
    if value is None:
        value = default_value

    return value


def get_bool_attribute(kwargs: dict, attribute: str, default_value: bool) -> bool:
    """returns value from kwargs whilst checking for bool type"""

    # getting attribute value
    value = get_attribute(kwargs, attribute, default_value)

    # checking for correct type
    if not isinstance(value, bool):
        raise ValueError(f"{attribute} should be a bool, not {type(value)}")

    return value



"""Tools to build simple buckets out of Qualitative features
for a binary classification model.
"""

from numpy import nan, select
from pandas import DataFrame, Series, unique

from ...features import BaseFeature, Features, GroupedList
from ...utils import extend_docstring
from ..utils.base_discretizer import BaseDiscretizer, Sample
from ..utils.type_discretizers import StringDiscretizer


class ChainedDiscretizer(BaseDiscretizer):
    """Automatic discretization of categorical features, joining rare modalities into higher
    level groups.

    For each provided :class:`GroupedList` from ``chained_orders`` attribute, values less frequent
    than ``min_freq`` are grouped in there respective group, as defined by :class:`GroupedList`.
    """

    __name__ = "ChainedDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__)
    def __init__(
        self,
        min_freq: float,
        features: list[BaseFeature],
        chained_orders: list[GroupedList],
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------

        chained_orders : list[GroupedList]
            A list of interlocked higher level groups for each modalities of each ordinal feature.
            Values of ``chained_orders[0]`` have to be grouped in ``chained_order[1]`` etc.
        """
        # not dropping nans whatsoever
        kwargs = dict(kwargs, dropna=False)
        super().__init__(features, **kwargs)  # Initiating BaseDiscretizer

        # class specific attributes
        self.min_freq = min_freq
        self.chained_orders = [GroupedList(values) for values in chained_orders]

        # known_values: all ordered values describe in each level of the chained_orders
        # starting off with first level
        known_values = self.chained_orders[0].values
        # adding each level
        for n, next_level in enumerate(self.chained_orders[1:]):
            # iterating over each group of the next level
            for next_group, next_values in next_level.content.items():
                # looking for known and unknwon values in next_level

                # checking for unknown values
                next_unknown = [
                    value
                    for value in next_values
                    if value not in known_values and value != next_group
                ]
                if len(next_unknown) > 0:
                    raise ValueError(
                        f"[{self.__name__}] Values {str(next_unknown)}, provided in "
                        f"chained_orders[{n+1}] are missing from chained_orders[{n}]. Please make "
                        "sure values are kept trhough each level."
                    )

                # checking for known values
                next_known = [
                    value for value in next_values if value in known_values and value != next_group
                ]
                if len(next_known) == 0:
                    raise ValueError(
                        f"[{self.__name__}] For key '{next_group}', the provided chained_orders"
                        f"[{n+1}] has no values from chained_orders[:{n+1}]. Please provide some"
                        " existing values."
                    )

                # index of the highest ranked known value of the next_group
                highest_index = known_values.index(next_known[-1])

                # adding next_group to the order at the right place in the amongst known_values
                known_values = (
                    known_values[: highest_index + 1]
                    + [next_group]
                    + known_values[highest_index + 1 :]
                )

        # saving resulting known values
        self.known_values = known_values

        # adding known_values to each feature's order
        for feature in self.features:
            # checking for already known values of the feature
            order = feature.values
            # no known values for the feature
            if order is None:
                order = GroupedList([])

            # checking that all values from the order are in known_values
            for value in order:
                if value not in self.known_values:
                    raise ValueError(
                        f"[{self.__name__}] Value {value} from feature {feature} provided in "
                        "values_orders is missing from levels of chained_orders. Add value to a "
                        "level of chained_orders or adapt values_orders."
                    )
            # adding known values if missing from the order
            for value in self.known_values:
                if value not in order.values:
                    order.append(value)
            # sorting in case an ordering was provided
            order = order.sort_by(self.known_values)
            # updating feature
            feature.update(order, replace=True)

    def _prepare_data(self, sample: Sample) -> Sample:
        """Validates format and content of X and y. Converts non-string columns into strings.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``ChainedDiscretizer.features``.

        y : Series
            Binary target feature, not used, by default None.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        # copying dataframe
        sample.X = sample.X.copy()

        # checking for binary target and previous fit
        sample = super()._prepare_data(sample)

        # checking feature values' frequencies
        check_frequencies(self.features, sample.X, self.min_freq, self.__name__)

        # converting non-str columns
        sample.X = ensure_qualitative_dtypes(self.features, sample.X, **self.kwargs)

        # fitting features
        self.features.fit(**sample)

        # filling up nans for features that have some
        sample.X = self.features.fillna(sample.X)

        # checking for unexpected values
        self.features.check_values(sample.X)

        return sample

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series = None) -> None:  # pylint: disable=W0222
        # preprocessing data
        sample = self._prepare_data(Sample(X, y))
        self._log_if_verbose()  # verbose if requested

        # iterating over each feature
        for feature in self.features:
            # iterating over each specified orders
            for level_order in self.chained_orders:
                # computing frequencies of each modality
                frequencies = (
                    sample.X[feature.version]
                    .value_counts(normalize=True, dropna=False)
                    .drop(nan, errors="ignore")
                )
                values, frequencies = frequencies.index, frequencies.values

                # values that are frequent enough
                to_keep = list(values[frequencies >= self.min_freq])

                # values from the order to group (not frequent enough or absent)
                values_to_group = [value for value in level_order.values if value not in to_keep]

                # values to group into discarded values
                groups_value = [level_order.get_group(value) for value in values_to_group]

                # values of the feature to input (needed for next levels of the order)
                df_to_input = [
                    sample.X[feature.version] == discarded for discarded in values_to_group
                ]

                # inputing non frequent values
                sample.X[feature.version] = select(
                    df_to_input, groups_value, default=sample.X[feature.version]
                )

                # historizing in the feature's order
                order = GroupedList(feature.values)
                for discarded, kept in zip(values_to_group, groups_value):
                    order.group(discarded, kept)

                # updating feature accordingly
                feature.update(order, replace=True)

        super().fit(X, y)

        if self.verbose:  # verbose if requested
            print("\n")

        return self


def check_frequencies(features: Features, X: DataFrame, min_freq: float, name: str) -> None:
    """Checks features' frequencies compared to min_freq"""

    # computing features' max modality frequency (mode frequency)
    max_freq = X[features.versions].apply(
        lambda u: u.value_counts(normalize=True, dropna=False).max(),
        axis=0,
    )

    # features with no common modality (biggest value less frequent than min_freq)
    non_common = [f.version for f in features if max_freq[f.version] < min_freq]

    # features with too common modality (biggest value more frequent than 1-min_freq)
    too_common = [f.version for f in features if max_freq[f.version] > 1 - min_freq]

    # raising
    if len(too_common + non_common) > 0:
        # building error message
        error_msg = (
            f"[{name}] Features {str(too_common + non_common)} contain a too frequent modality "
            "or no frequent enough modalities. Consider decreasing min_freq or removing these "
            "features.\nINFO:\n"
        )

        # adding features with no common values
        non_common = [
            (
                f" - {features(feature)}: most frequent value has "
                f"freq={max_freq[feature]:2.2%} < min_freq={min_freq:2.2%}"
            )
            for feature in non_common
        ]

        # adding features with too common values
        too_common = [
            (
                f" - {features(feature)}: most frequent value has "
                f"freq={max_freq[feature]:2.2%} > (1-min_freq)={1-min_freq:2.2%}"
            )
            for feature in too_common
        ]
        error_msg += "\n".join(too_common + non_common)

        raise ValueError(error_msg)


def ensure_qualitative_dtypes(features: Features, X: DataFrame, **kwargs) -> DataFrame:
    """Checks features' data types and converts int/float to str"""

    # getting per feature data types
    dtypes = (
        X.fillna({feature.version: feature.nan for feature in features})[features.versions]
        .map(type)
        .apply(unique, result_type="reduce")
    )

    # identifying features that are not str
    not_object = dtypes.apply(lambda u: any(dtype != str for dtype in u))

    # converting detected non-string features
    if any(not_object):
        # converting non-str features into qualitative features
        to_convert = [
            feature for feature in features if feature.version in not_object.index[not_object]
        ]
        string_discretizer = StringDiscretizer(features=to_convert, **kwargs)
        X = string_discretizer.fit_transform(X)

    return X



""" Defines a categorical feature"""

from ..utils.base_feature import BaseFeature
from .qualitative_feature import QualitativeFeature


class CategoricalFeature(QualitativeFeature):
    """Defines a categorical feature"""

    __name__ = "Categorical"
    is_categorical = True

    def _specific_formatting(self, ordered_content: list[str]) -> str:
        """categorical features' specific label formatting"""

        # list label for categorical feature
        label = ", ".join(ordered_content)
        if len(label) > self.max_n_chars:
            label = label[: self.max_n_chars] + "..."

        return label


def get_categorical_features(features: list[BaseFeature]) -> list[CategoricalFeature]:
    """returns categorical features amongst provided features"""
    return [feature for feature in features if feature.is_categorical]



"""Tools to build simple buckets out of Qualitative features
for a binary classification model.
"""

from pandas import DataFrame, Series, notna

from ...features import CategoricalFeature
from ...utils import extend_docstring
from ..utils.base_discretizer import BaseDiscretizer, Sample


class CategoricalDiscretizer(BaseDiscretizer):
    """Automatic discretization of categorical features, building simple groups frequent enough.

    Groups a qualitative features' values less frequent than ``min_freq`` into a ``str_default``
    string.

    NaNs are left untouched.

    Only use for qualitative non-ordinal features.
    """

    __name__ = "CategoricalDiscretizer"

    @extend_docstring(BaseDiscretizer.__init__, append=False, exclude=["features"])
    def __init__(
        self,
        categoricals: list[CategoricalFeature],
        min_freq: float,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        categoricals : list[CategoricalFeature]
            Categorical features to process
        """
        # Initiating BaseDiscretizer
        super().__init__(categoricals, **dict(kwargs, min_freq=min_freq))

    def _prepare_data(self, sample: Sample) -> Sample:
        """Validates format and content of X and y.

        Parameters
        ----------
        X : DataFrame
            Dataset used to discretize. Needs to have columns has specified in
            ``CategoricalDiscretizer.features``.

        y : Series
            Binary target feature with wich the association is maximized.

        Returns
        -------
        DataFrame
            A formatted copy of X
        """
        # checking for binary target
        sample = super()._prepare_data(sample)

        # fitting features
        self.features.fit(**sample)

        # filling up nans for features that have some
        sample.X = self.features.fillna(sample.X)

        return sample

    @extend_docstring(BaseDiscretizer.fit)
    def fit(self, X: DataFrame, y: Series) -> None:  # pylint: disable=W0222
        # copying dataframe and checking data before bucketization
        sample = self._prepare_data(Sample(X, y))

        self._log_if_verbose()  # verbose if requested

        # grouping modalities less frequent than min_freq into feature.default
        sample.X = self._group_rare_modalities(sample.X)

        # sorting features' values by target rate
        self._target_sort(**sample)

        # discretizing features based on each feature's values_order
        super().fit(**sample)

        return self

    def _group_feature_rare_modalities(
        self, feature: CategoricalFeature, X: DataFrame, frequencies: DataFrame
    ) -> DataFrame:
        """Groups modalities less frequent than min_freq into feature.default"""

        # checking for rare values
        values_to_group = [
            value
            for value, freq in frequencies[feature.version].items()
            if freq < self.min_freq and value != feature.nan and notna(value)
        ]

        # checking for completly missing values (no frequency observed in X)
        missing_values = [
            value for value in feature.values if value not in frequencies[feature.version]
        ]
        if len(missing_values) > 0:
            raise ValueError(f"[{self.__name__}] Unexpected values {missing_values} for {feature}.")

        # grouping values to str_default if any
        if any(values_to_group):
            # adding default value to the order
            feature.has_default = True

            # grouping rare values in default value
            feature.group(values_to_group, feature.default)
            X.loc[X[feature.version].isin(values_to_group), feature.version] = feature.default

        return X

    def _group_rare_modalities(self, X: DataFrame) -> DataFrame:
        """Groups modalities less frequent than min_freq into feature.default"""

        # computing frequencies of each modality
        frequencies = X[self.features.versions].apply(series_value_counts, axis=0)

        # grouping rare modalities for each feature
        for feature in self.features:
            X = self._group_feature_rare_modalities(feature, X, frequencies)

        return X

    def _target_sort(self, X: DataFrame, y: Series) -> None:
        """Sorts features' values by target rate"""

        # computing target rate per modality for ordering
        target_rates = X[self.features.versions].apply(series_target_rate, y=y, axis=0)

        # sorting features' values based on target rates
        self.features.update(
            {feature: list(sorted_values) for feature, sorted_values in target_rates.items()},
            sorted_values=True,
        )


def series_target_rate(x: Series, y: Series, dropna: bool = True, ascending=True) -> dict:
    """Target y rate per modality of x into a dictionnary"""

    rates = y.groupby(x, dropna=dropna).mean().sort_index().sort_values(ascending=ascending)

    return rates.to_dict()


def series_value_counts(x: Series, dropna: bool = False, normalize: bool = True) -> dict:
    """Counts the values of each modality of a series into a dictionnary"""

    values = x.value_counts(dropna=dropna, normalize=normalize)

    return values.to_dict()



""" Filters based on association measures between Quantitative features.
"""

from numpy import ones, triu
from pandas import DataFrame

from ...features import BaseFeature, get_versions
from ...utils.extend_docstring import extend_docstring
from .base_filters import BaseFilter

# from statsmodels.stats.outliers_influence import variance_inflation_factor


# TODO
# def vif_filter(X: DataFrame, ranks: DataFrame, **params) -> dict[str, Any]:
#     """Computes Variance Inflation Factor (multicolinearity)

#     Parameters
#     ----------
#     thresh_vif, float: default inf
#       Maximum VIF between features
#     """

#     # accessing the prefered order
#     prefered_order = ranks.index

#     # initiating list association per feature
#     associations = []

#     # list of dropped features
#     dropped = []

#     # iterating over each column
#     for i, feature in enumerate(prefered_order):
#         # identifying remaining more associated features
#         better_features = [f for f in prefered_order[: i + 1] if f not in dropped]

#         X_vif = X[better_features]  # keeping only better features
#         X_vif = X_vif.dropna(axis=0)  # dropping NaNs for OLS

#         # computation of VIF
#         vif = nan
#         if len(better_features) > 1 and len(X_vif) > 0:
#             vif = variance_inflation_factor(X_vif.values, len(better_features) - 1)

#         # dropping the feature if it was too correlated to a better feature
#         if vif > params.get("thresh_vif", inf) and notna(vif):
#             dropped += [feature]

#         # kept feature: updating associations with this feature
#         else:
#             associations += [{"feature": feature, "vif_filter": vif}]

#     # formatting ouput to DataFrame
#     associations = DataFrame(associations).set_index("feature")

#     # applying filter on association
#     associations = ranks.join(associations, how="right")

#     return associations


class QuantitativeFilter(BaseFilter):
    """Computes max association between X and X (quantitative) excluding features
    that are correlated to a feature more associated with the target
    (defined by the ranks).
    """

    __name__ = "QuantitativeFilter"

    is_x_quantitative = True
    is_absolute = True

    @extend_docstring(BaseFilter.filter)
    def filter(self, X: DataFrame, ranks: list[BaseFeature]) -> list[BaseFeature]:
        # computing correlation between features
        X_corr = self._compute_correlation(X, ranks)

        # filtering too correlated features
        return self._filter_correlated_features(X_corr, ranks)

    def _compute_correlation(self, X: DataFrame, rank: list[BaseFeature]) -> DataFrame:
        """Computing correlation between features"""
        # absolute correlation between features
        X_corr = X[get_versions(rank)].corr(self.measure)

        # getting upper right part of the correlation matrix and removing autocorrelation
        return X_corr.where(triu(ones(X_corr.shape), k=1).astype(bool))

    def _filter_correlated_features(
        self, X_corr: DataFrame, ranks: list[BaseFeature]
    ) -> list[BaseFeature]:
        """filtering out features too correlated with a better ranked feature"""

        # iterating over each feature by target association order
        filtered: list[BaseFeature] = []
        for feature in ranks:
            # maximum correlation with a better feature
            correlation_with, worst_correlation = self._compute_worst_correlation(X_corr, feature)

            # checking for too much correlation
            valid = self._validate(feature, worst_correlation, correlation_with)

            # dropping feature if it was too correlated
            if not valid:
                X_corr.drop(feature.version, axis=0, inplace=True)
                X_corr.drop(feature.version, axis=1, inplace=True)

            # keeping feature
            else:
                filtered += [feature]

        return filtered

    def _compute_worst_correlation(
        self, X_corr: DataFrame, feature: BaseFeature
    ) -> tuple[str, float]:
        """Computes correlation with better features (filtering out X_corr)"""

        # correlation with more associated features
        corr_with_better_features = X_corr.loc[: feature.version, feature.version].fillna(0)

        # worst/maximum absolute correlation with better features
        return corr_with_better_features.agg(
            [lambda x: x.abs().idxmax(), lambda x: max(x.min(), x.max(), key=abs)]
        )

    def _validate(
        self, feature: BaseFeature, worst_correlation: float, correlation_with: str
    ) -> bool:
        """Checks if the worst correlation of a feature is above specified threshold"""
        # dropping the feature if it was too correlated to a better feature
        valid = True
        if abs(worst_correlation) > self.threshold:
            valid = False

        # update feature accordingly (update stats)
        self._update_feature(
            feature,
            worst_correlation,
            valid,
            info={
                "correlation_with": (
                    correlation_with if correlation_with != feature.version else "itself"
                )
            },
        )

        return valid


class SpearmanFilter(QuantitativeFilter):
    """Computes maximum Spearman's rho between quantitative features of ``X``"""

    __name__ = "SpearmanFilter"

    @extend_docstring(QuantitativeFilter.__init__)
    def __init__(self, threshold: float = 1.0) -> None:
        super().__init__(threshold)
        self.measure = "spearman"


class PearsonFilter(QuantitativeFilter):
    """Computes maximum Pearson's r between quantitative features of ``X``"""

    __name__ = "PearsonFilter"

    @extend_docstring(QuantitativeFilter.__init__)
    def __init__(self, threshold: float = 1.0) -> None:
        super().__init__(threshold)
        self.measure = "pearson"



""" Defines a datetime feature"""

from .quantitative_feature import QuantitativeFeature


class DatetimeFeature(QuantitativeFeature):
    """TODO"""

    def __init__(self, name: str, reference_date: str, **kwargs) -> None:
        super().__init__(name, **kwargs)
        self.reference_date = reference_date  # date of reference to compare with



""" Pretty print functions for selectors """

from pandas import DataFrame

from ...features import BaseFeature


def format_measure(feature: BaseFeature, measure: dict) -> dict:
    """Format measure for display"""
    # formatting measure
    formatted = {}
    for measure_name, measure_value in measure.items():
        # adding feature name and measure value
        formatted.update({"feature": feature, measure_name: measure_value.get("value")})
        # adding correlation with
        if not measure_value.get("info", {}).get("is_default"):
            correlation_with = measure_value.get("info", {}).get("correlation_with")
            # formatting target correlation
            if correlation_with != "target" and correlation_with is not None:
                formatted.update({measure_name.replace("Filter", "With"): correlation_with})

    return formatted


def format_ranked_features(features: list[BaseFeature]) -> DataFrame:
    """formats ranked features for display"""

    # adding measures and filters
    measures = []
    for feature in features:
        measures.append(format_measure(feature, {**feature.measures, **feature.filters}))

    # finding a sorting measure
    if len(measures) > 0:
        ranks = [col for measure in measures for col in measure.keys() if col.endswith("Rank")]
        if len(ranks) > 0:
            sort_by = ranks[0]
            return DataFrame(measures).sort_values(by=sort_by, ascending=True)
    return DataFrame(measures)


def prettier_measures(association: DataFrame) -> str:
    """Pretty display of measures and filters per feature on the same line"""

    # finding columns with indicators to colorize
    subset = [
        column
        for column in association.columns
        # checking for an association indicator
        if any(column.endswith(indic) for indic in ["Measure", "Filter"])
    ]

    # adding coolwarm color gradient
    nicer_association = association.style.background_gradient(cmap="coolwarm", subset=subset)
    # printing inline notebook
    nicer_association = nicer_association.set_table_attributes("style='display:inline'")

    # finding columns with indicators to colorize
    subset = [
        column
        for column in association.columns
        # checking for an association indicator
        if any(column.endswith(indic) for indic in ["Measure", "Filter"])
    ]

    # adding numerical columns
    subset += association.select_dtypes(include=["number"]).columns.tolist()

    # lower precision for specific columns
    nicer_association = nicer_association.format({measure: "{:.4f}" for measure in subset})

    # conversion to html
    return nicer_association._repr_html_()



""" Filters based on association measure between features and a binary target.
"""

from abc import ABC, abstractmethod

from pandas import DataFrame

from ...features import BaseFeature


class BaseFilter(ABC):
    """Base filter class."""

    __name__ = "BaseFilter"

    is_x_quantitative = False
    """ wether x is quantitative or not """

    is_x_qualitative = False
    """ wether x is qualitative or not """

    # info
    is_default = False
    """ wether the filter is an association measure or not """

    higher_is_better = False
    """ wether higher values are better or not """

    is_absolute = False
    """ wether the measure needs absolute value for comparison or not """

    def __init__(self, threshold: float = 1.0) -> None:
        """
        Parameters
        ----------
        threshold : float, optional
            Maximum threshold to reach, by default ``1.0``
        """

        self.measure = None
        self.threshold = threshold

    def __repr__(self) -> str:
        return self.__name__

    @abstractmethod
    def filter(self, X: DataFrame, ranks: list[BaseFeature]) -> list[BaseFeature]:
        """Filters out ranked features that reach the association threshold

        Parameters
        ----------
        X : DataFrame
            DataFrame containing features
        ranks : list[BaseFeature]
            List of ranked features to filter, from most to least associated

        Returns
        -------
        list[BaseFeature]
            Filtered list of features
        """

    def _update_feature(
        self,
        feature: BaseFeature,
        value: float,
        valid: bool,
        info: dict,
    ) -> None:
        """adds measure to specified feature"""

        # filter state info
        filter_state = {
            self.__name__: {
                "value": value,
                "threshold": self.threshold,
                "valid": valid,
                "info": {"higher_is_better": self.higher_is_better, **info},
            }
        }

        # updating statistics of the feature accordingly
        feature.filters.update(filter_state)


class ValidFilter(BaseFilter):
    """Valid filter class."""

    __name__ = "Valid"

    is_default = True
    is_x_quantitative = True
    is_x_qualitative = True

    def filter(self, X: DataFrame, ranks: list[BaseFeature]) -> list[BaseFeature]:
        """filters out all non-valid features fril ranks"""
        _ = X  # not used

        # iterating over each feature
        filtered = []
        for feature in ranks:
            # checking for non-valid measures, keeping feature
            if len(feature.measures) == 0 or all(
                measure.get("valid") for measure in feature.measures.values()
            ):
                filtered += [feature]

        return filtered


class NonDefaultValidFilter(ValidFilter):
    """Valid filter class for non-default metrics."""

    __name__ = "NonDefaultValid"

    is_default = False



""" Loads association measures."""

from .base_measures import AbsoluteMeasure, BaseMeasure, ModeMeasure, NanMeasure, OutlierMeasure
from .qualitative_measures import Chi2Measure, CramervMeasure, TschuprowtMeasure
from .quantitative_measures import (
    DistanceMeasure,
    IqrOutlierMeasure,
    KruskalMeasure,
    PearsonMeasure,
    ReversibleMeasure,
    RMeasure,
    SpearmanMeasure,
    ZscoreOutlierMeasure,
)

__all__ = [
    "AbsoluteMeasure",
    "NanMeasure",
    "ModeMeasure",
    "BaseMeasure",
    "OutlierMeasure",
    "Chi2Measure",
    "CramervMeasure",
    "PearsonMeasure",
    "SpearmanMeasure",
    "TschuprowtMeasure",
    "RMeasure",
    "DistanceMeasure",
    "IqrOutlierMeasure",
    "KruskalMeasure",
    "ZscoreOutlierMeasure",
    "ReversibleMeasure",
]



"""This module contains the utility functions for the combinations module."""

from .combination_evaluator import CombinationEvaluator
from .target_rate import TargetRate

__all__ = ["CombinationEvaluator", "TargetRate"]



""" Base measures that defines Quantitative and Qualitative features.
"""

from abc import ABC, abstractmethod

from pandas import Series, isnull, notna

from ...features import BaseFeature


class BaseMeasure(ABC):
    """Base measure of association between x and y"""

    __name__ = "BaseMeasure"

    is_x_quantitative = False
    """ wether x is quantitative or not """

    is_x_qualitative = False
    """ wether x is qualitative or not """

    is_y_qualitative = False
    """ wether y is qualitative or not """

    is_y_quantitative = False
    """ wether y is quantitative or not """

    is_y_binary = False
    """ wether y is binary or not """

    is_sortable = True
    """ wether the metric can be sorted or not """

    # info
    is_default = False
    """ wether the measure is an association measure or not """

    higher_is_better = True
    """ wether higher values are better or not """

    correlation_with = "target"
    """ info about correlation with which other feature """

    is_absolute = False
    """ wether the measure needs absolute value for comparison or not """

    is_reversible = False
    """ wether the measure's input can be reversed depending on there type or not """

    def __init__(self, threshold: float = 0.0) -> None:
        """
        Parameters
        ----------
        threshold : float, optional
            Minimum threshold to reach, by default ``0.0``
        """
        self.threshold = threshold
        self.value = None
        self._info = {}

    def __repr__(self) -> str:
        return self.__name__

    @abstractmethod
    def compute_association(self, x: Series, y: Series) -> float:
        """Computes association measure between ``x`` and ``y``

        Parameters
        ----------
        x : Series
            Feature
        y : Series
            Target feature

        Returns
        -------
        float
            Measure of association between ``x`` and ``y``
        """

    @property
    def info(self) -> dict:
        """gives info about the measure"""

        return dict(
            self._info,
            # adding default info about the measure
            higher_is_better=self.higher_is_better,
            correlation_with=self.correlation_with,
            is_default=self.is_default,
        )

    @info.setter
    def info(self, content: dict) -> None:
        """updates info"""
        self._info.update(content)

    def validate(self) -> bool:
        """Checks if :attr:`threshold` is reached

        Returns
        -------
        bool
            Whether the test is passed or not
        """

        if not isnull(self.value) and notna(self.value):
            return self.value >= self.threshold
        return False

    def to_dict(self) -> dict:
        """converts to a dict"""
        return {
            self.__name__: {
                "value": self.value,
                "threshold": self.threshold,
                "valid": self.validate(),
                "info": self.info,
            }
        }

    def _update_feature(self, feature: BaseFeature) -> None:
        """adds measure to specified feature"""

        # checking for a value
        if self.value is None:
            raise ValueError(f"[{self}] Use compute_association first!")

        # updating statistics of the feature accordingly
        feature.measures.update(self.to_dict())

    def reverse_xy(self) -> bool:
        """reverses values of x and y in compute_association"""

        # when its not implemented
        return False


class AbsoluteMeasure(BaseMeasure):
    """Absolute measure of association between x and y"""

    is_absolute = True

    def validate(self) -> bool:
        """Checks if :attr:`threshold` is reached

        Returns
        -------
        bool
            Whether the test is passed or not
        """
        if not isnull(self.value) and notna(self.value):
            return abs(self.value) >= self.threshold
        return False


class OutlierMeasure(BaseMeasure):
    """Outlier measure of association for a Quantitative feature"""

    is_default = True
    is_x_quantitative = True
    is_x_qualitative = False
    is_sortable = False

    # info
    higher_is_better = False
    correlation_with = "itself"

    def __init__(self, threshold: float = 1.0) -> None:
        """
        Parameters
        ----------
        threshold : float, optional
            Maximum threshold to reach, by default ``1.0``
        """
        super().__init__(threshold)

    @abstractmethod
    def compute_association(self, x: Series, y: Series = None) -> float:
        """Computes outlier measure on ``x``

        Parameters
        ----------
        x : Series
            Feature
        y : Series, optional
            Target feature, by default ``None``

        Returns
        -------
        float
            Measure of outliers on ``x``
        """

    def validate(self) -> bool:
        """Checks if :attr:`threshold` is reached

        Returns
        -------
        bool
            Whether the test is passed or not
        """
        if not isnull(self.value) and notna(self.value):
            return self.value < self.threshold
        return True


class NanMeasure(OutlierMeasure):
    """Measure of the percentage of NaNs"""

    __name__ = "Nan"
    is_x_quantitative = True
    is_x_qualitative = True

    def compute_association(self, x: Series, y: Series = None) -> float:
        """Computes frequency of ``nan`` in ``x``

        Parameters
        ----------
        x : Series
            Feature
        y : Series, optional
            Target feature, by default ``None``

        Returns
        -------
        float
            Measure of ``nan`` in ``x``
        """
        _ = y
        self.value = (x.isna() | x.isnull()).mean()
        return self.value


class ModeMeasure(OutlierMeasure):
    """Measure of the percentage of the mode"""

    __name__ = "Mode"
    is_x_quantitative = True
    is_x_qualitative = True

    def compute_association(self, x: Series, y: Series = None) -> float:
        """Computes frequency of ``x``'s mode

        Parameters
        ----------
        x : Series
            Feature
        y : Series, optional
            Target feature, by default ``None``

        Returns
        -------
        float
            Measure of ``x``'s mode
        """
        _ = y
        mode = x.mode(dropna=True).values[0]  # computing mode
        self.value = (x == mode).mean()  # Computing percentage of the mode
        return self.value



""" Module for binary combination evaluators. """

from abc import ABC

from numpy import add, array, searchsorted, sqrt, unique, zeros
from pandas import DataFrame, notna
from scipy.stats import chi2_contingency

from ..utils.combination_evaluator import AggregatedSample, CombinationEvaluator
from .binary_target_rates import BinaryTargetRate, TargetMean


class BinaryCombinationEvaluator(CombinationEvaluator, ABC):
    """Binary combination evaluator class."""

    is_y_binary = True

    def _init_target_rate(self, target_rate: BinaryTargetRate) -> None:
        """Initializes target rate."""
        if target_rate is None:
            self.target_rate = TargetMean()
        elif not isinstance(target_rate, BinaryTargetRate):
            raise ValueError("target_rate must be a BinaryTargetRate")
        else:
            self.target_rate = target_rate

    def _association_measure(
        self, xagg: AggregatedSample, n_obs: int = None, tol: float = 1e-10
    ) -> dict[str, float]:
        """Computes measures of association between feature and target by crosstab.

        Parameters
        ----------
        xtab : DataFrame
            Crosstab between feature and target.

        n_obs : int
            Sample total size.

        Returns
        -------
        dict[str, float]
            Cramér's V and Tschuprow's as a dict.
        """
        # number of values taken by the features
        n_mod_x = xagg.shape[0]

        # Chi2 statistic
        chi2 = chi2_contingency(xagg.values + tol)[0]

        # Cramér's V
        cramerv = sqrt(chi2 / n_obs)
        if notna(cramerv):
            cramerv = round(cramerv / tol) * tol

        # Tschuprow's T
        tschuprowt = cramerv / sqrt(sqrt(n_mod_x - 1))
        if notna(tschuprowt):
            tschuprowt = round(tschuprowt / tol) * tol

        return {"cramerv": cramerv, "tschuprowt": tschuprowt}

    def _grouper(self, xagg: AggregatedSample, groupby: dict) -> DataFrame:
        """Groups a crosstab by groupby and sums column values by groups (vectorized)

        Parameters
        ----------
        xagg : DataFrame
            crosstab between X and y
        groupby : list[str]
            indices to group by

        Returns
        -------
        DataFrame
            Crosstab grouped by indices
        """
        # all indices that may be duplicated
        index_values = array([groupby.get(index_value, index_value) for index_value in xagg.index])

        # all unique indices deduplicated
        unique_indices = unique(index_values)

        # initiating summed up array with zeros
        summed_values = zeros((len(unique_indices), len(xagg.columns)))

        # for each unique_index found in index_values sums xtab.Values at corresponding position
        # in summed_values
        add.at(summed_values, searchsorted(unique_indices, index_values), xagg.values)

        # converting back to dataframe
        return DataFrame(summed_values, index=unique_indices, columns=xagg.columns)


class TschuprowtCombinations(BinaryCombinationEvaluator):
    """Tschuprow's T based combination evaluation toolkit"""

    sort_by = "tschuprowt"


class CramervCombinations(BinaryCombinationEvaluator):
    """Cramér's V based combination evaluation toolkit"""

    sort_by = "cramerv"



""" Loads Carving tools."""

from .carvers.binary_carver import BinaryCarver
from .carvers.continuous_carver import ContinuousCarver
from .carvers.multiclass_carver import MulticlassCarver
from .features import Features
from .selectors import ClassificationSelector, RegressionSelector

__all__ = [
    "BinaryCarver",
    "ContinuousCarver",
    "Features",
    "MulticlassCarver",
    "ClassificationSelector",
    "RegressionSelector",
]
